# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as _dpg

_dpg.create_context()

from deareis.data import Project
from deareis.api.data import (
    DataSet,
    # - exceptions
    UnsupportedFileFormat,
    # - functions
    parse_data,
)
from deareis.api.circuit import (
    Circuit,
    CircuitBuilder,
    # - connections
    Connection,
    Parallel,
    Series,
    # - elements
    Element,
    Capacitor,
    ConstantPhaseElement,
    DeLevieFiniteLength,
    Gerischer,
    HavriliakNegami,
    HavriliakNegamiAlternative,
    Inductor,
    ModifiedInductor,
    Resistor,
    Warburg,
    WarburgOpen,
    WarburgShort,
    # - exceptions
    ParsingError,
    UnexpectedCharacter,
    # - functions
    get_elements,
    parse_cdc,
)
from deareis.api.kramers_kronig import (
    TestResult,
    TestSettings,
    # - enums
    TestMode,
    Test,
    # - functions
    perform_test,
    perform_exploratory_tests,
)
from deareis.api.fitting import (
    FitResult,
    FitSettings,
    FittedParameter,
    # - enums
    CNLSMethod,
    Weight,
    # - exceptions
    FittingError,
    # - functions
    fit_circuit,
)
from deareis.api.simulation import (
    SimulationResult,
    SimulationSettings,
    # - functions
    simulate_spectrum,
)
from deareis.api.plotting import (
    PlotSeries,
    PlotSettings,
    # - enums
    PlotType,
)
from deareis.api.drt import (
    DRTResult,
    DRTSettings,
    # - enums
    DRTMethod,
    DRTMode,
    RBFShape,
    RBFType,
    # - exceptions
    DRTError,
    # - functions
    calculate_drt,
)
from deareis.api.plot import mpl  # matplotlib-based plotting



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


boukamp1995: str = "DOI:10.1149/1.2044210"
schonleber2014: str = "DOI:10.1016/j.electacta.2014.01.034"

kramers_kronig = SimpleNamespace(
    **{
        "test": f"""
Fit an equivalent circuit (Voigt circuit, fig. 1) to an impedance spectrum.

CNLS: Use complex non-linear least squares fitting.

Complex: Use a set of linear equations to fit to the complex impedance (eq. 12).

Imaginary: Use a set of linear equations to fit to the imaginary part of the complex impedance (eqs. 4 and 7).

Real: Use a set of linear equations to fit to the real part of the complex impedance (eqs. 8 and 10).

Reference: {boukamp1995}
    """.strip(),
        "mode": f"""
Auto: Increment the number of RC elements, calculate the µ-value (eq. 21), and use the µ-criterion to determine when to stop.

Exploratory: Test multiple numbers of RC elements similarly to the 'Auto' setting, show all the results, and let the user choose one.

Manual: Choose a specific number of RC elements to use.

Reference: {schonleber2014}
    """.strip(),
        "max_num_RC": """
The (maximum) number of RC elements connected in series in the equivalent circuit.

WARNING! If the test is set to 'CNLS' and the mode is set to 'Exploratory', then it may take a long time to perform the test. It is recommended that you initially lower this setting to e.g. thirty if there are more measured frequencies than that in your data set.
    """.strip(),
        "add_capacitance": """
Add a capacitor in series to the equivalent circuit. This may be necessary in some circumstances such as if the imaginary part of the spectrum does not approach zero when the frequency is decreased.
    """.strip(),
        "add_inductance": """
Add an inductor in series to the equivalent circuit. This may be necessary in some circumstances and is enabled by default for all but the 'CNLS' test.
    """.strip(),
        "mu_criterion": f"""
The µ-value represents how the equivalent circuit fits the data (eq. 21). µ-values range from zero to one with the extremes representing over-fitting and under-fitting, respectively. Over-fitting is undesirable since the equivalent circuit could end up reproducing any noise that may be present in the data. However, too much under-fitting means that the equivalent circuit is unable to reproduce the data.

The µ-criterion defines the threshold at which no more RC elements are added to the equivalent circuit. Note that in some cases the calculated µ-value may prematurely and temporarily drop below the threshold set by the µ-criterion when performing the test using the 'Auto' mode.

The 'Exploratory' mode is recommended since you can inspect the µ-value produced by various other equivalent circuits and see if the number of RC elements by the 'Auto' mode was erroneous. Note that the number of RC elements initially suggested by the 'Exploratory' mode also takes into account factors other than the µ-value. Thus, the two modes may suggest different values for the number of RC elements.

This setting is only used when the mode setting is set to either 'Auto' or 'Exploratory'.

Reference: {schonleber2014}
    """.strip(),
        "method": """
The iterative method used to perform the fitting.

This setting is only used when the test setting is set to 'CNLS'.
    """.strip(),
        "nfev": """
The maximum number of function evaluations to use when fitting. This can be used to limit the amount of time spent performing a fit. A value of zero means that there is no limit.

This setting is only used when the test setting is set to 'CNLS'.
    """.strip(),
        "perform": f"""
The purpose of the Kramers-Kronig test is to help validate the linearity and time-invariance of impedance spectra. Multiple variants of the linear Kramers-Kronig test are implemented in this program. This test involves fitting an equivalent circuit, which is known to be Kramers-Kronig transformable, to an impedance spectrum.

The default settings (i.e. 'Complex' test and 'Exploratory' mode) are recommended at least as a starting point.

Valid spectra typically exhibit:
- Randomly distributed residuals that are centered along the x-axis.
- Residuals with small magnitudes (e.g. less than half of a percent). Note that the residuals of noisy impedance spectra may have a greater magnitude and still be valid.

Invalid spectra typically exhibit:
- Residuals that are not randomly distributed. Note that if the number of RC elements used to perform the test is too small, then the residuals may form a sinusoidal pattern.
- Residuals that are clearly biased away from the x-axis. For example, residuals may increasinly diverge from the x-axis as the frequency is decreased or they may form a convex/concave shape.

See the following references for more information about the linear Kramers-Kronig tests:

- {boukamp1995}

- {schonleber2014}

Analysing the contributions of higher-order harmonics to the response signal is another way of checking for non-linearity. Potentiostat/galvanostat manufacturers may implement this in different ways and use different terminology may be used in their documentation. For example, "total harmonic distortion" (THD) may be available as a setting in the measurement method. Alternatively, it may be possible to record the response signal for processing and analysis (e.g., Fourier transform).
    """.strip(),
        "delete": """
Delete the current test result.
    """.strip(),
        "pseudo_chisqr": f"""
Pseudo chi-squared value calculated according to eq. 14.

Reference: {boukamp1995}
    """.strip(),
        "exploratory_result": """
The result to highlight and ultimately save.
    """.strip(),
        "mu": f"""
The µ-value represents how the equivalent circuit fits the data (eq. 21). µ-values range from zero to one with the extremes representing over-fitting and under-fitting, respectively.

Reference: {schonleber2014}
    """.strip(),
        "num_RC": """
The number of RC elements connected in series in the equivalent circuit.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from datetime import datetime
from hashlib import sha1
from os.path import exists
from typing import (
    List,
    Optional,
    Set,
    Tuple,
    Union,
)
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg
import dearpygui.dearpygui as dpg
from numpy import (
    all as array_all,
    asarray,
    float32,
    floating,
    floor,
    inf,
    integer,
    isclose,
    issubdtype,
    log10 as log,
    nan,
    ndarray,
    pad,
)


MATH_REGISTRY: int = dpg.add_texture_registry()


def calculate_checksum(*args, **kwargs) -> str:
    string: str = kwargs.get("string", "")
    path: str = kwargs.get("path", "")
    assert type(string) is str, string
    assert type(path) is str, path
    assert string != "" or (path != "" and exists(path)), (
        string,
        path,
    )
    checksum: str
    if string:
        checksum = sha1(string.encode()).hexdigest()
    elif path:
        with open(path, "rb") as fp:
            checksum = sha1(fp.read()).hexdigest()
    return checksum


def calculate_window_position_dimensions(
    width: Union[int, float] = 0.9, height: Union[int, float] = 0.9
) -> Tuple[int, int, int, int]:
    assert (issubdtype(type(width), floating) and width > 0.0 and width < 1.0) or (
        issubdtype(type(width), integer) and width > 0
    )
    assert (issubdtype(type(height), floating) and height > 0.0 and height < 1.0) or (
        issubdtype(type(width), integer) and height > 0
    )
    viewport_width: int = dpg.get_viewport_width()
    x: int
    if issubdtype(type(width), floating):
        x = floor(viewport_width * (1.0 - width) / 2)
        width = floor(viewport_width * width)
    else:
        x = floor((viewport_width - width) / 2)
    viewport_height: int = dpg.get_viewport_height()
    y: int
    if issubdtype(type(height), floating):
        y = floor(viewport_height * (1.0 - height) / 2)
        height = floor(viewport_height * height)
    else:
        y = floor((viewport_height - height) / 2)
    return (
        int(x),
        int(y),
        int(width),
        int(height),
    )


def format_timestamp(timestamp: float) -> str:
    assert issubdtype(type(timestamp), floating), timestamp
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")


def format_number(
    value: float,
    decimals: int = 1,
    width: int = -1,
    exponent: bool = True,
    significants: int = 0,
) -> str:
    float(value)  # Make sure that the value can at least be converted to a float
    assert issubdtype(type(decimals), integer), decimals
    assert issubdtype(type(width), integer), width
    assert type(exponent) is bool, exponent
    assert issubdtype(type(significants), integer) and significants >= 0, significants
    fmt: str = "{:." + str(decimals) + "f}"
    if significants > 0:
        fmt = "{:." + str(significants) + "g}"
    string: str
    if value == nan or value is nan:
        string = "NaN"
    elif value == inf:
        string = "INF"
    elif value == -inf:
        string = "-INF"
    elif exponent:
        if value == 0.0:
            string = "0"
        else:
            exp: int = int(floor(log(abs(value)) / 3) * 3)
            coeff: float = value / 10**exp
            string = fmt.format(coeff).lower()
            if "e+03" in string:
                string = string[: string.find("e")]
                exp += 3
            if exp >= 0:
                string += "e+{:02d}".format(exp)
            else:
                string += "e-{:02d}".format(abs(exp))
    else:
        string = fmt.format(value)
    if significants > 0:
        i: int = string.find("e") if "e" in string else len(string)
        if "." not in string and i < significants:
            string = (string[:i] + ".").ljust(significants + 1, "0") + string[i:]
        elif "." in string and i < significants + 1:
            string = string[:i].ljust(significants + 1, "0") + string[i:]
    if width > 1:
        string = string.rjust(width)
    if string.endswith("e+00"):
        string = string.replace("e+00", "    ")
    return string


def align_numbers(values: List[str]) -> List[str]:
    assert type(values) is list and all(map(lambda _: type(_) is str, values))
    has_negative_values: bool = any(map(lambda _: _.strip().startswith("-"), values))
    values = list(
        map(
            lambda _: _
            if not has_negative_values
            else ((" " if not _.startswith("-") else "") + _),
            map(str.strip, values),
        )
    )
    indices: List[int] = list(
        map(
            lambda _: max(
                _.find("."),
                _.lower().find("e") if "." not in _ else 0,
                len(_) if "." not in _ and "e" not in _.lower() else 0,
            ),
            values,
        )
    )
    max_index: int = max(indices)
    for i, index in enumerate(indices):
        value = values[i]
        values[i] = value.rjust(max_index - index + len(value))
    return values


def pad_dataframe_dictionary(dictionary: dict) -> Optional[dict]:
    lengths: Set[int] = set(map(len, dictionary.values()))
    max_len: int = max(lengths)
    if max_len == 0:
        return None
    elif len(lengths) > 1:
        padded_dictionary: dict = dictionary.copy()
        for label in padded_dictionary.keys():
            padded_dictionary[label] = pad(
                padded_dictionary[label],
                (
                    0,
                    max_len - len(padded_dictionary[label]),
                ),
                constant_values=None,
            )
        return padded_dictionary
    return dictionary


def is_filtered_item_visible(item: int, filter_string: str) -> bool:
    filter_string = filter_string.strip()
    if filter_string == "":
        return True
    filter_key: Optional[str] = dpg.get_item_filter_key(item)
    if filter_key is None:
        return False
    visible: bool = False
    for fragment in map(str.strip, filter_string.split(",")):
        if fragment.startswith("-"):
            visible = True
            if fragment[1:] in filter_key:
                return False
        else:
            visible = False
            if fragment in filter_key:
                return True
    return visible


def render_math(
    math: str,
    width: int,
    height: int,
    dpi: int = 100,
    fontsize: int = 12,
    alpha: float = 0.0,
    print_dimensions: bool = False,
) -> int:
    fig: Figure = plt.figure(
        figsize=(
            width / dpi,
            height / dpi,
        ),
        dpi=dpi,
    )
    canvas: FigureCanvasAgg = FigureCanvasAgg(fig)
    fig.patch.set_alpha(alpha)
    ax: Axes = fig.gca()
    ax.text(
        0.0,
        0.5,
        math,
        va="center",
        color="white",
        fontsize=fontsize,
    )
    ax.set_axis_off()
    fig.subplots_adjust(
        left=0.0,
        bottom=0.0,
        right=1.0,
        top=1.0,
        wspace=0.0,
        hspace=0.0,
    )
    canvas.draw()
    buffer: ndarray = asarray(canvas.buffer_rgba()).astype(float32) / 255
    if print_dimensions:
        min_x: int = 0
        max_x: int = buffer.shape[1] - 1
        min_y: int = 0
        max_y: int = buffer.shape[0] - 1
        i: int
        row: ndarray
        col: ndarray
        rgba_sums: ndarray
        for i, row in enumerate(buffer):
            rgba_sums = row.sum(1)
            if isclose(rgba_sums[0], 3.0 + alpha, atol=1e-2) and array_all(
                isclose(rgba_sums, rgba_sums[0])
            ):
                # All pixels in this row are just the background (1.0, 1.0, 1.0, alpha)
                min_y = i
            else:
                break
        for (i, row) in reversed(list(enumerate(buffer))):
            rgba_sums = row.sum(1)
            if isclose(rgba_sums[0], 3.0 + alpha, atol=1e-2) and array_all(
                isclose(rgba_sums, rgba_sums[0])
            ):
                # All pixels in this row are just the background (1.0, 1.0, 1.0, alpha)
                max_y = i + 1
            else:
                break
        for i in range(buffer.shape[1]):
            col = buffer[:, i]
            rgba_sums = col.sum(1)
            if isclose(rgba_sums[0], 3.0 + alpha, atol=1e-2) and array_all(
                isclose(rgba_sums, rgba_sums[0])
            ):
                # All pixels in this column are just the background (1.0, 1.0, 1.0, alpha)
                min_x = i
            else:
                break
        for i in reversed(range(buffer.shape[1])):
            col = buffer[:, i]
            rgba_sums = col.sum(1)
            if isclose(rgba_sums[0], 3.0 + alpha, atol=1e-2) and array_all(
                isclose(rgba_sums, rgba_sums[0])
            ):
                # All pixels in this column are just the background (1.0, 1.0, 1.0, alpha)
                max_x = i + 1
            else:
                break
        print(f"{math}\n- width: {max_x - min_x}\n- height: {max_y - min_y}")
    tag: int = dpg.add_raw_texture(
        buffer.shape[1],
        buffer.shape[0],
        buffer,
        format=dpg.mvFormat_Float_rgba,
        parent=MATH_REGISTRY,
    )
    plt.close(fig)
    return tag



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Optional
import dearpygui.dearpygui as dpg
from pyimpspec import Circuit
from deareis.gui.circuit_editor.parser import Parser


class CircuitPreview:
    def __init__(self, width: int = -1, height: int = -1):
        self.node_editor: int = dpg.generate_uuid()
        dpg.add_node_editor(
            width=width,
            height=height,
            tag=self.node_editor,
            minimap=True,
            minimap_location=dpg.mvNodeMiniMap_Location_BottomRight,
        )
        self.parser: Parser = Parser(self.node_editor)

    def clear(self):
        dpg.delete_item(self.node_editor, children_only=True)

    def update(self, circuit: Optional[Circuit]):
        self.clear()
        if circuit is None:
            return
        self.parser.circuit_to_nodes(circuit)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from json import (
    dumps as dump_json,
    loads as parse_json,
)
from os.path import (
    dirname,
    exists,
)
from traceback import format_exc
from typing import (
    Dict,
    List,
    Optional,
)
import dearpygui.dearpygui as dpg
from deareis.data import (
    DataSet,
    PlotSettings,
    Project,
    SimulationResult,
)
from deareis.enums import Context
from deareis.gui import ProjectTab
from deareis.gui.file_dialog import FileDialog
from deareis.signals import Signal
from deareis.state import STATE
from deareis.utility import calculate_window_position_dimensions
import deareis.signals as signals


def new_project(*args, **kwargs):
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Creating new project")
    project: Project = Project()
    project_tab: ProjectTab
    existing_tab: bool
    project_tab, existing_tab = STATE.add_project(project)
    STATE.program_window.select_tab(project_tab)
    signals.emit(Signal.SELECT_PROJECT_TAB, uuid=project.uuid)
    if not existing_tab:
        STATE.snapshot_project_state(project)
        signals.emit(
            Signal.RESTORE_PROJECT_STATE,
            project=project,
            project_tab=project_tab,
            state_snapshot="{}",
        )
        assert STATE.is_project_dirty(project) is True
    paths: List[str] = kwargs.get("data", [])
    if paths:
        signals.emit(Signal.LOAD_DATA_SET_FILES, paths=paths)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def select_project_files(*args, **kwargs):
    merge: bool = kwargs.get("merge", False)
    # Check if any recent projects have been selected in the home tab
    paths: List[str] = STATE.program_window.get_selected_projects()
    if paths:
        signals.emit(
            Signal.LOAD_PROJECT_FILES,
            paths=paths,
            merge=merge,
        )
        return
    FileDialog(
        cwd=STATE.latest_project_directory,
        label="Select project file(s)",
        callback=lambda *a, **k: signals.emit(
            Signal.LOAD_PROJECT_FILES,
            **k,
        ),
        extensions=[".json"],
        merge=merge,
    )


def load_project_files(*args, **kwargs):
    paths: List[str] = kwargs.get("paths", [])
    merge: bool = kwargs.get("merge", False)
    assert type(paths) is list, paths
    assert type(merge) is bool, merge
    project: Project
    project_tab: ProjectTab
    existing_tab: bool
    assert paths, len(paths)
    parsing_errors: Dict[str, str] = {}
    path: str
    assert len(paths) >= 1, paths
    if merge:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Merging projects")
        projects: List[Project] = []
        for path in paths:
            try:
                projects.append(Project.from_file(path))
            except Exception:
                parsing_errors[path] = format_exc()
                continue
        if not parsing_errors:
            project = Project.merge(projects)
            project_tab, existing_tab = STATE.add_project(project)
            STATE.program_window.select_tab(project_tab)
            signals.emit(Signal.SELECT_PROJECT_TAB, uuid=project.uuid)
            if not existing_tab:
                STATE.snapshot_project_state(project)
                signals.emit(
                    Signal.RESTORE_PROJECT_STATE,
                    project=project,
                    project_tab=project_tab,
                    state_snapshot=dump_json(project.to_dict(session=True)),
                )
                project_tab.set_dirty(STATE.is_project_dirty(project))
                assert STATE.is_project_dirty(project) is True
        STATE.set_recent_projects(paths=[])
    else:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading project(s)")
        for path in paths:
            try:
                project: Project = Project.from_file(path)
            except Exception:
                parsing_errors[path] = format_exc()
                continue
            project_tab, existing_tab = STATE.add_project(project)
            STATE.program_window.select_tab(project_tab)
            signals.emit(Signal.SELECT_PROJECT_TAB, uuid=project.uuid)
            if not existing_tab:
                STATE.snapshot_project_state(project)
                signals.emit(
                    Signal.RESTORE_PROJECT_STATE,
                    project=project,
                    project_tab=project_tab,
                    state_snapshot=dump_json(project.to_dict(session=True)),
                )
                STATE.update_project_state_saved_index(project)
                project_tab.set_dirty(STATE.is_project_dirty(project))
                assert STATE.is_project_dirty(project) is False
                assert (
                    STATE.get_project_state_snapshot_index(project) == 0
                ), STATE.get_project_state_snapshot_index(project)
        STATE.set_recent_projects(
            paths=list(filter(lambda _: _ not in parsing_errors, paths))
        )
    STATE.latest_project_directory = dirname(path)
    if parsing_errors:
        total_traceback: str = ""
        traceback: str
        for path, traceback in parsing_errors.items():
            total_traceback += f"{traceback}\nThe exception above was encountered while parsing '{path}'.\n\n"
        signals.emit(
            Signal.SHOW_ERROR_MESSAGE,
            traceback=total_traceback.strip(),
            message="""
Encountered error(s) while parsing project file(s). The file(s) might be malformed, corrupted, or simply a newer version than is supported by this version of DearEIS.
            """.strip(),
        )
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def restore_project_state(*args, **kwargs):
    project: Optional[Project] = kwargs.get("project", STATE.get_active_project())
    project_tab: Optional[ProjectTab] = kwargs.get(
        "project_tab", STATE.get_active_project_tab()
    )
    state_snapshot: Optional[str] = kwargs.get("state_snapshot")
    if project is None or project_tab is None or state_snapshot is None:
        return
    project_state: dict = parse_json(state_snapshot)
    project.update(**project_state)
    project_tab.set_label(project_state.get("label", "Project"))
    project_tab.set_notes(project_state.get("notes", ""))
    project_tab.populate_plots(project)
    plot: Optional[PlotSettings] = project_tab.get_active_plot()
    plots: List[PlotSettings] = project.get_plots()
    if plot is None or plot.uuid not in list(map(lambda _: _.uuid, plots)):
        assert len(plots) > 0
        plot = plots[0]
    else:
        # The PlotSettings returned by ProjectTab.get_active_plot is potentially out of date.
        # Look for the up-to-date version among those returned by Project.get_plots.
        plot = list(filter(lambda _: _.uuid == plot.uuid, plots))[0]
    signals.emit(
        Signal.SELECT_PLOT_SETTINGS,
        settings=plot,
    )
    project_tab.populate_data_sets(project)
    data: Optional[DataSet] = project_tab.get_active_data_set(
        context=Context.DATA_SETS_TAB
    )
    data_sets: List[DataSet] = project.get_data_sets()
    if data is None and data_sets:
        data = data_sets[0]
    elif data_sets:
        # This is done because the DataSet instance returned by get_active_data_set is outdated.
        found_data: bool = False
        for _ in data_sets:
            if _.uuid == data.uuid:
                data = _
                found_data = True
                break
        if not found_data:
            data = data_sets[0]
    else:
        data = None
    signals.emit(
        Signal.SELECT_DATA_SET,
        data=data,
    )
    project_tab.populate_simulations(project)
    simulation: Optional[SimulationResult] = project_tab.get_active_simulation()
    simulations: List[SimulationResult] = project.get_simulations()
    if simulation is None and simulations:
        simulation = simulations[0]
    elif simulations:
        found_sim: bool = False
        for _ in simulations:
            if _.uuid == simulation.uuid:
                simulation = _
                found_sim = True
                break
        if not found_sim:
            simulation = simulations[0]
    else:
        simulation = None
    signals.emit(
        Signal.SELECT_SIMULATION_RESULT,
        simulation=simulation,
        data=project_tab.get_active_data_set(context=Context.SIMULATION_TAB),
    )
    project_tab.set_dirty(STATE.is_project_dirty(project))


def create_project_snapshot(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    STATE.snapshot_project_state(project)
    project_tab.set_dirty(kwargs.get("dirty", True))


def save_project_as(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    FileDialog(
        cwd=STATE.latest_project_directory,
        label="Select project path",
        callback=lambda *a, **k: signals.emit(
            Signal.SAVE_PROJECT,
            close_project=kwargs.get("close_project", False),
            **k,
        ),
        extensions=[".json"],
        save=True,
    )


def save_project(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    path: Optional[str] = kwargs.get("path", "").strip() or None
    if path is None and project.get_path().strip() == "":
        signals.emit(
            Signal.SAVE_PROJECT_AS, close_project=kwargs.get("close_project", False)
        )
        return
    assert path is None or exists(
        dirname(path)
    ), f"Folder does not exist: '{dirname(path)}'"
    project.save(path)
    STATE.update_project_state_saved_index(project)
    project_tab.set_dirty(STATE.is_project_dirty(project))
    STATE.set_recent_projects(paths=[project.get_path()])
    STATE.clear_project_backups([project])
    if kwargs.get("close_project", False):
        STATE.remove_project(project)
        dpg.delete_item(project_tab.tab)


def close_project(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    if not STATE.is_project_dirty(project) or kwargs.get("force", False):
        STATE.remove_project(project)
        dpg.delete_item(project_tab.tab)
        STATE.clear_project_backups([project])
        return
    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)

    def discard():
        close()
        STATE.remove_project(project)
        dpg.delete_item(project_tab.tab)
        STATE.clear_project_backups([project])

    def save():
        close()
        signals.emit(Signal.SAVE_PROJECT, close_project=True)

    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(300, 56)
    with dpg.window(
        label="Unsaved changes",
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_resize=True,
        modal=True,
        tag=window,
    ):
        dpg.add_text(
            "Do you wish to save the changes, discard the changes, or cancel?",
            wrap=260,
        )
        dpg.add_spacer(height=8)
        with dpg.group(horizontal=True):
            dpg.add_button(
                label="Save",
                callback=save,
            )
            dpg.add_button(label="Discard", callback=discard)
            dpg.add_button(label="Cancel", callback=close)
    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close,
        )
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from json import dumps as dump_json
from os import (
    getcwd,
    makedirs,
    remove,
    walk,
)
from os.path import (
    exists,
    join,
)
from typing import (
    Any,
    Dict,
    IO,
    List,
    Optional,
    Tuple,
)
from xdg import (
    xdg_state_home,  # User-specific state data files
)
import dearpygui.dearpygui as dpg
from deareis.version import PACKAGE_VERSION
from deareis.config import Config
from deareis.data import (
    PlotSettings,
    Project,
)
from deareis.enums import (
    Context,
)
from deareis.gui.project import ProjectTab
from deareis.gui.program import ProgramWindow
from deareis.gui.command_palette import CommandPalette
from deareis.keybindings import KeybindingHandler
from deareis.utility import calculate_checksum
from deareis.gui.plotting.export import PlotExporter
from deareis.signals import Signal
import deareis.signals as signals


class State:
    def __init__(self):
        self.state_directory_path: str = join(xdg_state_home(), "DearEIS")
        if not exists(self.state_directory_path):
            makedirs(self.state_directory_path)
        self.snapshots_directory_path: str = join(
            self.state_directory_path, "snapshots"
        )
        if not exists(self.snapshots_directory_path):
            makedirs(self.snapshots_directory_path)
        self.recent_projects_path: str = join(
            self.state_directory_path, "recent_projects"
        )
        self.recent_projects: List[str] = []
        self.config: Config = Config()
        self.program_window: ProgramWindow = ProgramWindow()
        self.latest_project_directory: str = getcwd()
        self.latest_data_set_directory: str = getcwd()
        self.latest_plot_directory: str = getcwd()
        self.projects: List[Project] = []
        self.project_tabs: Dict[str, ProjectTab] = {}
        self.project_lookup: Dict[str, Tuple[Project, ProjectTab]] = {}
        self.active_project: Optional[Project] = None
        self.keybinding_handler: KeybindingHandler = KeybindingHandler(
            self.config.keybindings, self
        )
        self.active_modal_window: Optional[int] = None
        self.active_modal_window_object: Any = None
        self.project_state_snapshots: Dict[str, List[str]] = {}
        self.project_state_snapshot_indices: Dict[str, int] = {}
        self.project_state_saved_indices: Dict[str, int] = {}
        self.command_palette: CommandPalette = CommandPalette(self.keybinding_handler)
        self.plot_exporter: PlotExporter = PlotExporter(self.config)

    def set_active_modal_window(self, window: int, window_object: Any):
        assert type(window) is int
        if dpg.does_item_exist(window):
            self.active_modal_window = window
            self.active_modal_window_object = window_object
        else:
            self.active_modal_window = None
            self.active_modal_window_object = None

    def get_active_modal_window(self) -> Optional[int]:
        return self.active_modal_window

    def add_project(self, project: Project) -> Tuple[ProjectTab, bool]:
        existing_tab: bool = project.uuid in self.project_lookup
        if existing_tab:
            return (
                self.project_tabs[project.uuid],
                existing_tab,
            )
        project_tab: ProjectTab = ProjectTab(
            project.uuid,
            self.program_window.project_tab_bar.tab_bar,
            self,
        )
        self.projects.append(project)
        self.project_tabs[project.uuid] = project_tab
        self.project_lookup[project.uuid] = (
            project,
            project_tab,
        )
        return (
            project_tab,
            existing_tab,
        )

    def remove_project(self, project: Project):
        assert project in self.projects
        self.projects.remove(project)
        del self.project_tabs[project.uuid]
        del self.project_lookup[project.uuid]
        del self.project_state_snapshots[project.uuid]
        del self.project_state_snapshot_indices[project.uuid]
        del self.project_state_saved_indices[project.uuid]

    def set_active_project(self, uuid: Optional[str]):
        assert (type(uuid) is str and uuid.strip() != "") or uuid is None
        if uuid is None:
            self.active_project = None
        else:
            self.active_project = self.project_lookup[uuid][0]

    def get_active_project(self) -> Optional[Project]:
        return self.active_project

    def get_active_project_tab(self) -> Optional[ProjectTab]:
        if self.active_project is None:
            return None
        return self.project_tabs[self.active_project.uuid]

    def get_project_tab(self, project: Project) -> Optional[ProjectTab]:
        return self.project_tabs.get(project.uuid)

    def update_project_state_saved_index(self, project: Project):
        assert type(project) is Project, project
        index: int = self.project_state_snapshot_indices[project.uuid]
        self.project_state_saved_indices[project.uuid] = index

    def is_project_dirty(self, project: Project) -> bool:
        assert type(project) is Project, project
        index: int = self.project_state_saved_indices[project.uuid]
        return index < 0 or self.get_project_state_snapshot_index(project) != index

    def get_project_state_snapshot_index(self, project: Project) -> int:
        assert type(project) is Project, project
        return self.project_state_snapshot_indices[project.uuid]

    def snapshot_project_state(self, project: Project):
        assert type(project) is Project, project
        if project.uuid not in self.project_state_snapshots:
            self.project_state_snapshots[project.uuid] = []
            self.project_state_snapshot_indices[project.uuid] = 0
            self.project_state_saved_indices[project.uuid] = -1
        snapshots: List[str] = self.project_state_snapshots[project.uuid]
        index: int = self.get_project_state_snapshot_index(project)
        while len(snapshots) - 1 > index:
            snapshots.pop()
        if index < self.project_state_saved_indices[project.uuid]:
            self.project_state_saved_indices[project.uuid] = -1
        snapshots.append(
            dump_json(project.to_dict(session=True)),
        )
        self.project_state_snapshot_indices[project.uuid] = len(snapshots) - 1
        if (
            index > 0
            and self.config.auto_backup_interval > 0
            and index % self.config.auto_backup_interval == 0
        ):
            self.serialize_project_snapshots([project], auto_backup=True)

    def get_previous_project_state_snapshot(self, project: Project) -> Optional[str]:
        assert type(project) is Project, project
        if project.uuid not in self.project_state_snapshots:
            return None
        snapshots: List[str] = self.project_state_snapshots[project.uuid]
        if not snapshots:
            return None
        index: int = self.get_project_state_snapshot_index(project)
        if index < 1:
            return None
        self.project_state_snapshot_indices[project.uuid] = index - 1
        return snapshots[index - 1]

    def get_next_project_state_snapshot(self, project: Project) -> Optional[str]:
        assert type(project) is Project, project
        if project.uuid not in self.project_state_snapshots:
            return None
        snapshots: List[str] = self.project_state_snapshots[project.uuid]
        if not snapshots:
            return None
        index: int = self.get_project_state_snapshot_index(project)
        if index >= len(snapshots) - 1:
            return None
        self.project_state_snapshot_indices[project.uuid] = index + 1
        return snapshots[index + 1]

    def get_recent_projects(self) -> List[str]:
        if not self.recent_projects and exists(self.recent_projects_path):
            fp: IO
            with open(self.recent_projects_path, "r") as fp:
                self.recent_projects.extend(
                    list(filter(exists, map(str.strip, fp.readlines())))
                )
        return self.recent_projects.copy()

    def set_recent_projects(self, *args, **kwargs):
        paths: List[str] = kwargs.get("paths")
        assert type(paths) is list, paths
        for path in reversed(paths):
            if path in self.recent_projects:
                self.recent_projects.remove(path)
            self.recent_projects.insert(0, path)
        self.program_window.project_tab_bar.home_tab.update_recent_projects_table(
            self.recent_projects
        )

    def clear_recent_projects(self, selected_projects: List[str]):
        assert type(selected_projects) is list and all(
            map(lambda _: type(_) is str, selected_projects)
        ), selected_projects
        if len(selected_projects) > 0:
            for path in selected_projects:
                if path in self.recent_projects:
                    self.recent_projects.remove(path)
        else:
            self.recent_projects.clear()
        self.program_window.project_tab_bar.home_tab.update_recent_projects_table(
            self.recent_projects
        )

    def save_recent_projects(self):
        contents: str = "\n".join(
            list(filter(lambda _: exists(_), self.recent_projects))
        )
        if exists(self.recent_projects_path) and contents != "":
            new_checksum: str = calculate_checksum(string=contents)
            old_checksum: str = calculate_checksum(path=self.recent_projects_path)
            if new_checksum == old_checksum:
                return
        fp: IO
        with open(self.recent_projects_path, "w") as fp:
            fp.write(contents)

    def serialize_project_snapshots(
        self, projects: List[Project], auto_backup: bool = False
    ):
        project: Project
        for project in projects:
            snapshot: dict = project.to_dict(session=False)
            snapshot["path"] = project.get_path()
            path: str = join(
                self.snapshots_directory_path,
                f"{project.uuid}{'-auto-backup' if auto_backup else ''}.json",
            )
            fp: IO
            with open(path, "w") as fp:
                fp.write(dump_json(snapshot))
        if not auto_backup:
            self.clear_project_backups(projects)

    def get_unsaved_project_snapshots(self) -> List[str]:
        root: str
        paths: List[str]
        for root, _, paths in walk(self.snapshots_directory_path):
            break
        return list(
            sorted(
                map(
                    lambda _: join(root, _),
                    filter(lambda _: _.endswith(".json"), paths),
                )
            )
        )

    def clear_project_backups(self, projects: List[Project]):
        project: Project
        for project in projects:
            path: str = join(
                self.snapshots_directory_path, f"{project.uuid}-auto-backup.json"
            )
            if not exists(path):
                continue
            remove(path)

    def show_command_palette(self):
        project: Optional[Project] = self.get_active_project()
        project_tab: Optional[ProjectTab] = self.get_active_project_tab()
        contexts: List[Context] = [Context.PROGRAM]
        if project is not None and project_tab is not None:
            contexts.extend([Context.PROJECT, project_tab.get_active_context()])
        self.command_palette.show(contexts, project, project_tab)

    def show_plot_exporter(self, settings: PlotSettings, project: Project):
        self.plot_exporter.show(settings, project)

    def close_plot_exporter(self):
        self.plot_exporter.close()

    def is_busy_message_visible(self) -> bool:
        return self.program_window.busy_message.is_visible()

    def check_version(self):
        recent_version_path: str = join(self.state_directory_path, "recent_version")
        fp: IO
        if not exists(recent_version_path):
            with open(recent_version_path, "w") as fp:
                fp.write(PACKAGE_VERSION)
            return
        version: str = ""
        with open(recent_version_path, "r") as fp:
            version = fp.read().strip()
        if version != PACKAGE_VERSION:
            with open(recent_version_path, "w") as fp:
                fp.write(PACKAGE_VERSION)
            signals.emit(Signal.SHOW_CHANGELOG)


STATE: State = State()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    List,
    Optional,
    Tuple,
)
import dearpygui.dearpygui as dpg
from numpy import (
    array,
    ceil,
    floor,
    log10 as log,
    ndarray,
)
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class Bode(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=True,
                location=dpg.mvPlot_Location_North,
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="f (Hz)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis_1: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="|Z| (ohm)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis_2: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="-phi (°)",
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return (
            len(dpg.get_item_children(self._y_axis_1, slot=1)) == 0
            and len(dpg.get_item_children(self._y_axis_2, slot=1)) == 0
        )

    def clear(self, *args, **kwargs):
        delete: bool = kwargs.get("delete", True)
        if delete:
            dpg.delete_item(self._y_axis_1, children_only=True)
            dpg.delete_item(self._y_axis_2, children_only=True)
            self._series.clear()
        else:
            i: int
            series_1: int
            series_2: int
            for i, (series_1, series_2) in enumerate(
                zip(
                    dpg.get_item_children(self._y_axis_1, slot=1),
                    dpg.get_item_children(self._y_axis_2, slot=1),
                )
            ):
                self._series[i]["frequency"] = array([])
                self._series[i]["magnitude"] = array([])
                self._series[i]["phase"] = array([])
                dpg.set_value(series_1, [[], []])
                dpg.set_value(series_2, [[], []])

    def update(self, index: int, *args, **kwargs):
        assert type(index) is int and index >= 0, index
        assert len(self._series) > index, (
            index,
            len(self._series),
        )
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        mag: ndarray = kwargs["magnitude"]
        phase: ndarray = kwargs["phase"]
        assert type(freq) is ndarray, freq
        assert type(mag) is ndarray, mag
        assert type(phase) is ndarray, phase
        i: int
        series_1: int
        series_2: int
        for i, (series_1, series_2) in enumerate(
            zip(
                dpg.get_item_children(self._y_axis_1, slot=1),
                dpg.get_item_children(self._y_axis_2, slot=1),
            )
        ):
            if i != index:
                continue
            self._series[index].update(kwargs)
            dpg.set_value(series_1, [list(freq), list(mag)])
            dpg.set_value(series_2, [list(freq), list(phase)])
            break

    def plot(self, *args, **kwargs) -> Tuple[int, int]:
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        mag: ndarray = kwargs["magnitude"]
        phase: ndarray = kwargs["phase"]
        labels: Tuple[str, str] = kwargs["labels"]
        sim: bool = kwargs.get("simulation", False)
        fit: bool = kwargs.get("fit", False)
        line: bool = kwargs.get("line", False)
        show_labels: bool = kwargs.get("show_labels", True)
        themes: Optional[Tuple[int, int]] = kwargs.get("themes")
        assert type(freq) is ndarray, freq
        assert type(mag) is ndarray, mag
        assert type(phase) is ndarray, phase
        assert type(labels) is tuple and len(labels) == 2, labels
        assert type(sim) is bool, sim
        assert type(fit) is bool, fit
        assert type(line) is bool, line
        assert (type(themes) is tuple and len(themes) == 2) or themes is None, themes
        assert type(show_labels) is bool, show_labels
        self._series.append(kwargs)
        func: Callable = dpg.add_scatter_series if not line else dpg.add_line_series
        x: list = list(freq)
        tag_mag: int = func(
            x=x,
            y=list(mag),
            label=labels[0] if show_labels else None,
            parent=self._y_axis_1,
        )
        tag_phase: int = func(
            x=x,
            y=list(phase),
            label=labels[1] if show_labels else None,
            parent=self._y_axis_2,
        )
        if themes is not None:
            dpg.bind_item_theme(tag_mag, themes[0])
            dpg.bind_item_theme(tag_phase, themes[1])
        return (
            tag_mag,
            tag_phase,
        )

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        x_min: Optional[float] = None
        x_max: Optional[float] = None
        y1_min: Optional[float] = None
        y1_max: Optional[float] = None
        y2_min: Optional[float] = None
        y2_max: Optional[float] = None
        for kwargs in self._series:
            freq: ndarray = kwargs["frequency"]
            mag: ndarray = kwargs["magnitude"]
            phase: ndarray = kwargs["phase"]
            if freq.size > 0:
                if x_min is None or min(freq) < x_min:
                    x_min = min(freq)
                if x_max is None or max(freq) > x_max:
                    x_max = max(freq)
            if mag.size > 0:
                if y1_min is None or min(mag) < y1_min:
                    y1_min = min(mag)
                if y1_max is None or max(mag) > y1_max:
                    y1_max = max(mag)
            if phase.size > 0:
                if y2_min is None or min(phase) < y2_min:
                    y2_min = min(phase)
                if y2_max is None or max(phase) > y2_max:
                    y2_max = max(phase)
        if x_min is None:
            x_min = 0.0
            x_max = 1.0
            y1_min = 0.0
            y1_max = 1.0
            y2_min = 0.0
            y2_max = 1.0
        else:
            dx: float = 0.1
            x_min = 10 ** (floor(log(x_min) / dx) * dx - dx)
            x_max = 10 ** (ceil(log(x_max) / dx) * dx + dx)
            dy: float = 0.1
            y1_min = 10 ** (floor(log(y1_min) / dy) * dy - dy)
            y1_max = 10 ** (ceil(log(y1_max) / dy) * dy + dy)
            if log(y1_max) - log(y1_min) < 1.0:
                y1_min = 10 ** floor(log(y1_min))
                y1_max = 10 ** ceil(log(y1_max))
            n: int = 5
            y2_min = floor(y2_min / n) * n - n
            y2_max = ceil(y2_max / n) * n + n
        dpg.split_frame()
        dpg.set_axis_limits(self._x_axis, ymin=x_min, ymax=x_max)
        dpg.set_axis_limits(self._y_axis_1, ymin=y1_min, ymax=y1_max)
        dpg.set_axis_limits(self._y_axis_2, ymin=y2_min, ymax=y2_max)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis_1)
        dpg.set_axis_limits_auto(self._y_axis_2)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis_1,
                other._y_axis_2,
            ],
            [
                self._x_axis,
                self._y_axis_1,
                self._y_axis_2,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis_1)
        dpg.set_axis_limits_auto(self._y_axis_2)


class BodeMagnitude(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=kwargs.get("legend_horizontal", True),
                location=kwargs.get("legend_location", dpg.mvPlot_Location_North),
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="f (Hz)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="|Z| (ohm)",
                log_scale=True,
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return len(dpg.get_item_children(self._y_axis, slot=1)) == 0

    def clear(self, *args, **kwargs):
        dpg.delete_item(self._y_axis, children_only=True)
        self._series.clear()

    def plot(self, *args, **kwargs) -> int:
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        mag: ndarray = kwargs["magnitude"]
        label: str = kwargs["label"]
        sim: bool = kwargs.get("simulation", False)
        fit: bool = kwargs.get("fit", False)
        line: bool = kwargs.get("line", False)
        theme: Optional[int] = kwargs.get("theme")
        show_label: bool = kwargs.get("show_label", True)
        assert type(freq) is ndarray, freq
        assert type(mag) is ndarray, mag
        assert type(label) is str, label
        assert type(sim) is bool, sim
        assert type(fit) is bool, fit
        assert type(line) is bool, line
        assert type(theme) is int or theme is None, theme
        assert type(show_label) is bool, show_label
        self._series.append(kwargs)
        func: Callable = dpg.add_scatter_series if not line else dpg.add_line_series
        tag: int = func(
            x=list(freq),
            y=list(mag),
            label=label if show_label else None,
            parent=self._y_axis,
        )
        if theme is not None:
            dpg.bind_item_theme(tag, theme)
        return tag

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        x_min: Optional[float] = None
        x_max: Optional[float] = None
        y_min: Optional[float] = None
        y_max: Optional[float] = None
        for kwargs in self._series:
            freq: ndarray = kwargs["frequency"]
            mag: ndarray = kwargs["magnitude"]
            if freq.size > 0:
                if x_min is None or min(freq) < x_min:
                    x_min = min(freq)
                if x_max is None or max(freq) > x_max:
                    x_max = max(freq)
            if mag.size > 0:
                if y_min is None or min(mag) < y_min:
                    y_min = min(mag)
                if y_max is None or max(mag) > y_max:
                    y_max = max(mag)
        if x_min is None:
            x_min = 0.0
            x_max = 1.0
            y_min = 0.0
            y_max = 1.0
        else:
            dx: float = 0.1
            x_min = 10 ** (floor(log(x_min) / dx) * dx - dx)
            x_max = 10 ** (ceil(log(x_max) / dx) * dx + dx)
            dy: float = 0.1
            y_min = 10 ** (floor(log(y_min) / dy) * dy - dy)
            y_max = 10 ** (ceil(log(y_max) / dy) * dy + dy)
            if log(y_max) - log(y_min) < 1.0:
                y_min = 10 ** floor(log(y_min))
                y_max = 10 ** ceil(log(y_max))
        dpg.split_frame()
        dpg.set_axis_limits(self._x_axis, ymin=x_min, ymax=x_max)
        dpg.set_axis_limits(self._y_axis, ymin=y_min, ymax=y_max)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis,
            ],
            [
                self._x_axis,
                self._y_axis,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)


class BodePhase(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=kwargs.get("legend_horizontal", True),
                location=kwargs.get("legend_location", dpg.mvPlot_Location_North),
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="f (Hz)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="-phi (°)",
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return len(dpg.get_item_children(self._y_axis, slot=1)) == 0

    def clear(self, *args, **kwargs):
        dpg.delete_item(self._y_axis, children_only=True)
        self._series.clear()

    def plot(self, *args, **kwargs) -> int:
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        phase: ndarray = kwargs["phase"]
        label: str = kwargs["label"]
        sim: bool = kwargs.get("simulation", False)
        fit: bool = kwargs.get("fit", False)
        line: bool = kwargs.get("line", False)
        theme: Optional[int] = kwargs.get("theme")
        show_label: bool = kwargs.get("show_label", True)
        assert type(freq) is ndarray, freq
        assert type(phase) is ndarray, phase
        assert type(label) is str, label
        assert type(sim) is bool, sim
        assert type(fit) is bool, fit
        assert type(line) is bool, line
        assert type(theme) is int or theme is None, theme
        assert type(show_label) is bool, show_label
        self._series.append(kwargs)
        func: Callable = dpg.add_scatter_series if not line else dpg.add_line_series
        tag: int = func(
            x=list(freq),
            y=list(phase),
            label=label if show_label else None,
            parent=self._y_axis,
        )
        if theme is not None:
            dpg.bind_item_theme(tag, theme)
        return tag

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        x_min: Optional[float] = None
        x_max: Optional[float] = None
        y_min: Optional[float] = None
        y_max: Optional[float] = None
        for kwargs in self._series:
            freq: ndarray = kwargs["frequency"]
            phase: ndarray = kwargs["phase"]
            if freq.size > 0:
                if x_min is None or min(freq) < x_min:
                    x_min = min(freq)
                if x_max is None or max(freq) > x_max:
                    x_max = max(freq)
            if phase.size > 0:
                if y_min is None or min(phase) < y_min:
                    y_min = min(phase)
                if y_max is None or max(phase) > y_max:
                    y_max = max(phase)
        if x_min is None:
            x_min = 0.0
            x_max = 1.0
            y_min = 0.0
            y_max = 1.0
        else:
            dx: float = 0.1
            x_min = 10 ** (floor(log(x_min) / dx) * dx - dx)
            x_max = 10 ** (ceil(log(x_max) / dx) * dx + dx)
            n: int = 5
            y_min = floor(y_min / n) * n - n
            y_max = ceil(y_max / n) * n + n
        dpg.split_frame()
        dpg.set_axis_limits(self._x_axis, ymin=x_min, ymax=x_max)
        dpg.set_axis_limits(self._y_axis, ymin=y_min, ymax=y_max)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis,
            ],
            [
                self._x_axis,
                self._y_axis,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import basename
from typing import List, Optional
import webbrowser
import dearpygui.dearpygui as dpg
from deareis.gui.busy_message import BusyMessage
from deareis.gui.error_message import ErrorMessage
from deareis.gui.project import ProjectTab
from deareis.signals import Signal
import deareis.signals as signals
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips


class HomeTab:
    def __init__(self):
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(
            label="Home",
            order_mode=dpg.mvTabOrder_Fixed,
            tag=self.tab,
        ):
            dpg.add_text("Recent projects")
            self.recent_projects_table: int = dpg.generate_uuid()
            with dpg.child_window(border=False, width=-1, height=-24):
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    height=-1,
                    tag=self.recent_projects_table,
                ):
                    dpg.add_table_column(width_fixed=True)
                    dpg.add_table_column(label="Filename")
                    dpg.add_table_column(width_fixed=True)
            with dpg.child_window(border=False):
                with dpg.group(horizontal=True):
                    dpg.add_button(
                        label="New project",
                        callback=lambda: signals.emit(Signal.NEW_PROJECT),
                    )
                    attach_tooltip(tooltips.home.new_project)
                    self.load_projects_button: int = dpg.generate_uuid()
                    dpg.add_button(
                        label="Load project(s)",
                        callback=lambda: signals.emit(Signal.SELECT_PROJECT_FILES),
                        tag=self.load_projects_button,
                    )
                    attach_tooltip(tooltips.home.load_projects)
                    self.merge_projects_button: int = dpg.generate_uuid()
                    dpg.add_button(
                        label="Merge projects",
                        callback=lambda: signals.emit(
                            Signal.SELECT_PROJECT_FILES,
                            merge=True,
                        ),
                        tag=self.merge_projects_button,
                    )
                    attach_tooltip(tooltips.home.merge_projects)
                    self.clear_recent_projects_button: int = dpg.generate_uuid()
                    dpg.add_button(
                        label="Clear recent projects",
                        callback=lambda: signals.emit(
                            Signal.CLEAR_RECENT_PROJECTS,
                            selected_projects=self.get_selected_projects(),
                        ),
                        tag=self.clear_recent_projects_button,
                    )
                    attach_tooltip(tooltips.home.clear_recent_projects)

    def update_recent_projects_table(self, paths: List[str]):
        assert type(paths) is list, paths
        dpg.delete_item(self.recent_projects_table, children_only=True, slot=1)
        for path in paths:
            with dpg.table_row(
                parent=self.recent_projects_table,
            ):
                dpg.add_button(
                    label="Load",
                    callback=lambda s, a, u: signals.emit(
                        Signal.LOAD_PROJECT_FILES,
                        paths=[u],
                    ),
                    user_data=path,
                )
                attach_tooltip(tooltips.home.load)
                dpg.add_text(basename(path))
                attach_tooltip(path)
                dpg.add_checkbox(
                    user_data=path,
                    callback=lambda s, a, u: self.updated_selection(),
                )
                attach_tooltip(tooltips.recent_projects.checkbox)

    def updated_selection(self):
        paths: List[str] = self.get_selected_projects()
        if len(paths) > 1:
            dpg.set_item_label(
                self.load_projects_button,
                "Load selected projects",
            )
            dpg.set_item_label(
                self.merge_projects_button,
                "Merge selected projects",
            )
            dpg.set_item_label(
                self.clear_recent_projects_button,
                "Clear selected recent projects",
            )
        elif len(paths) == 1:
            dpg.set_item_label(
                self.load_projects_button,
                "Load selected project",
            )
            dpg.set_item_label(
                self.merge_projects_button,
                "Merge selected projects",
            )
            dpg.set_item_label(
                self.clear_recent_projects_button,
                "Clear selected recent project",
            )
        else:
            dpg.set_item_label(
                self.load_projects_button,
                "Load project(s)",
            )
            dpg.set_item_label(
                self.merge_projects_button,
                "Merge projects",
            )
            dpg.set_item_label(
                self.clear_recent_projects_button,
                "Clear recent projects",
            )

    def get_selected_projects(self) -> List[str]:
        paths: List[str] = []
        row: int
        for row in dpg.get_item_children(self.recent_projects_table, slot=1):
            checkbox: int = dpg.get_item_children(row, slot=1)[-2]
            assert "Checkbox" in dpg.get_item_type(checkbox), dpg.get_item_type(
                checkbox
            )
            if dpg.get_value(checkbox):
                path: Optional[str] = dpg.get_item_user_data(checkbox)
                assert type(path) is str, path
                paths.append(path)
        return paths


class ProjectTabBar:
    def __init__(self):
        self.tab_bar: int = dpg.generate_uuid()
        with dpg.tab_bar(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_PROJECT_TAB
                if dpg.get_item_user_data(a) is not None
                else Signal.SELECT_HOME_TAB,
                uuid=dpg.get_item_user_data(a),
            ),
            tag=self.tab_bar,
        ):
            self.home_tab: HomeTab = HomeTab()

    def select_tab(self, project_tab: ProjectTab):
        dpg.set_value(self.tab_bar, project_tab.tab)

    def select_home_tab(self):
        dpg.set_value(self.tab_bar, self.home_tab.tab)

    def select_next_tab(self):
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        index: int = tabs.index(dpg.get_value(self.tab_bar)) + 1
        dpg.set_value(self.tab_bar, tabs[index % len(tabs)])

    def select_previous_tab(self):
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        index: int = tabs.index(dpg.get_value(self.tab_bar)) - 1
        dpg.set_value(self.tab_bar, tabs[index % len(tabs)])


class MenuBar:
    def __init__(self):
        button: int
        with dpg.menu_bar():
            with dpg.menu(label="File"):
                dpg.add_menu_item(
                    label="New project",
                    callback=lambda: signals.emit(Signal.NEW_PROJECT),
                )
                dpg.add_menu_item(
                    label="Load project",
                    callback=lambda: signals.emit(Signal.SELECT_PROJECT_FILES),
                )
                button = dpg.add_menu_item(
                    label="Save project",
                    enabled=False,
                    callback=lambda: signals.emit(Signal.SAVE_PROJECT),
                )
                signals.register(
                    Signal.SELECT_HOME_TAB,
                    lambda b=button, *args, **kwargs: dpg.disable_item(b),
                )
                signals.register(
                    Signal.SELECT_PROJECT_TAB,
                    lambda b=button, *args, **kwargs: dpg.enable_item(b),
                )
                button = dpg.add_menu_item(
                    label="Save project as",
                    enabled=False,
                    callback=lambda: signals.emit(Signal.SAVE_PROJECT_AS),
                )
                signals.register(
                    Signal.SELECT_HOME_TAB,
                    lambda b=button, *args, **kwargs: dpg.disable_item(b),
                )
                signals.register(
                    Signal.SELECT_PROJECT_TAB,
                    lambda b=button, *args, **kwargs: dpg.enable_item(b),
                )
                button = dpg.add_menu_item(
                    label="Close project",
                    enabled=False,
                    callback=lambda: signals.emit(Signal.CLOSE_PROJECT),
                )
                signals.register(
                    Signal.SELECT_HOME_TAB,
                    lambda b=button, *args, **kwargs: dpg.disable_item(b),
                )
                signals.register(
                    Signal.SELECT_PROJECT_TAB,
                    lambda b=button, *args, **kwargs: dpg.enable_item(b),
                )
                dpg.add_menu_item(label="Exit", callback=dpg.stop_dearpygui)
            with dpg.menu(label="Edit"):
                dpg.add_menu_item(
                    label="Undo",
                    enabled=False,
                    callback=lambda s, a, u: signals.emit(Signal.UNDO_PROJECT_ACTION),
                )
                button = dpg.last_item()
                signals.register(
                    Signal.SELECT_HOME_TAB,
                    lambda b=button, *args, **kwargs: dpg.disable_item(b),
                )
                signals.register(
                    Signal.SELECT_PROJECT_TAB,
                    lambda b=button, *args, **kwargs: dpg.enable_item(b),
                )
                dpg.add_menu_item(
                    label="Redo",
                    enabled=False,
                    callback=lambda s, a, u: signals.emit(Signal.REDO_PROJECT_ACTION),
                )
                button = dpg.last_item()
                signals.register(
                    Signal.SELECT_HOME_TAB,
                    lambda b=button, *args, **kwargs: dpg.disable_item(b),
                )
                signals.register(
                    Signal.SELECT_PROJECT_TAB,
                    lambda b=button, *args, **kwargs: dpg.enable_item(b),
                )
            with dpg.menu(label="Settings"):
                dpg.add_menu_item(
                    label="Appearance",
                    callback=lambda: signals.emit(Signal.SHOW_SETTINGS_APPEARANCE),
                )
                dpg.add_menu_item(
                    label="Defaults",
                    callback=lambda: signals.emit(Signal.SHOW_SETTINGS_DEFAULTS),
                )
                dpg.add_menu_item(
                    label="Keybindings",
                    callback=lambda: signals.emit(Signal.SHOW_SETTINGS_KEYBINDINGS),
                )
            with dpg.menu(label="Help"):
                dpg.add_menu_item(
                    label="Tutorials",
                    callback=lambda: webbrowser.open(
                        "https://vyrjana.github.io/DearEIS/tutorials/"
                    ),
                )
                dpg.add_menu_item(
                    label="About", callback=lambda: signals.emit(Signal.SHOW_HELP_ABOUT)
                )
                dpg.add_menu_item(
                    label="Changelog",
                    callback=lambda: signals.emit(Signal.SHOW_CHANGELOG),
                )
                dpg.add_menu_item(
                    label="Check for updates",
                    callback=lambda: signals.emit(Signal.CHECK_UPDATES),
                )
                dpg.add_menu_item(
                    label="Licenses",
                    callback=lambda: signals.emit(Signal.SHOW_HELP_LICENSES),
                )


class ProgramWindow:
    def __init__(self):
        self.window: int = dpg.generate_uuid()
        with dpg.window(tag=self.window):
            self.menu_bar: MenuBar = MenuBar()
            self.project_tab_bar: ProjectTabBar = ProjectTabBar()
        self.busy_message: BusyMessage = BusyMessage()
        self.error_message: ErrorMessage = ErrorMessage()

    def select_tab(self, project_tab: ProjectTab):
        self.project_tab_bar.select_tab(project_tab)

    def select_home_tab(self):
        self.project_tab_bar.select_home_tab()

    def select_next_tab(self):
        self.project_tab_bar.select_next_tab()

    def select_previous_tab(self):
        self.project_tab_bar.select_previous_tab()

    def get_selected_projects(self) -> List[str]:
        return self.project_tab_bar.home_tab.get_selected_projects()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from collections import OrderedDict
from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    List,
    Tuple,
    Union,
)
from numpy import (
    angle,
    array,
    integer,
    issubdtype,
    ndarray,
)
from pandas import DataFrame
import pyimpspec
from pyimpspec import (
    Circuit,
    Element,
    FittedParameter,
)
from pyimpspec.analysis.fitting import _interpolate
from deareis.enums import (
    CNLSMethod,
    Weight,
)
from deareis.utility import format_timestamp


VERSION: int = 1


def _parse_settings_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "cdc": dictionary["cdc"],
        "method": CNLSMethod(dictionary["method"]),
        "weight": Weight(dictionary["weight"]),
        "max_nfev": dictionary["max_nfev"],
    }


@dataclass(frozen=True)
class FitSettings:
    """
    A class to store the settings used to perform a circuit fit.

    Parameters
    ----------
    cdc: str
        The circuit description code (CDC) for the circuit to fit.

    method: CNLSMethod
        The iterative method to use when performing the fit.

    weight: Weight
        The weight function to use when performing the fit.

    max_nfev: int
        The maximum number of function evaluations to use when performing the fit.
    """

    cdc: str
    method: CNLSMethod
    weight: Weight
    max_nfev: int

    def __repr__(self) -> str:
        return f"FitSettings ({hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "FitSettings":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_settings_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        return Class(**parsers[version](dictionary))

    def to_dict(self) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        return {
            "version": VERSION,
            "cdc": self.cdc,
            "method": self.method,
            "weight": self.weight,
            "max_nfev": self.max_nfev,
        }


def _parse_result_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "uuid": dictionary["uuid"],
        "timestamp": dictionary["timestamp"],
        "circuit": pyimpspec.parse_cdc(dictionary["circuit"]),
        "parameters": {
            element_label: {
                parameter_label: FittedParameter.from_dict(param)
                for parameter_label, param in parameters.items()
            }
            for element_label, parameters in dictionary["parameters"].items()
        },
        "frequency": array(dictionary["frequency"]),
        "impedance": array(
            list(
                map(
                    lambda _: complex(*_),
                    zip(
                        dictionary["real_impedance"],
                        dictionary["imaginary_impedance"],
                    ),
                )
            )
        ),
        "mask": {int(k): v for k, v in dictionary.get("mask", {}).items()},
        "real_residual": array(dictionary["real_residual"]),
        "imaginary_residual": array(dictionary["imaginary_residual"]),
        "chisqr": dictionary["chisqr"],
        "red_chisqr": dictionary["red_chisqr"],
        "aic": dictionary["aic"],
        "bic": dictionary["bic"],
        "ndata": dictionary["ndata"],
        "nfree": dictionary["nfree"],
        "nfev": dictionary["nfev"],
        "method": dictionary["method"],
        "weight": dictionary["weight"],
        "settings": FitSettings.from_dict(dictionary["settings"]),
    }


@dataclass
class FitResult:
    """
    A class containing the result of a circuit fit.

    Parameters
    ----------
    uuid: str
        The universally unique identifier assigned to this result.

    timestamp: float
        The Unix time (in seconds) for when the test was performed.

    circuit: Circuit
        The final, fitted circuit.

    parameters: Dict[str, Dict[str, FittedParameter]]
        The mapping to the mappings of the final, fitted values of the element parameters.

    frequency: ndarray
        The frequencies used to perform the fit.

    impedance: ndarray
        The complex impedances of the fitted circuit at each of the frequencies.

    real_residual: ndarray
        The residuals of the real part of the complex impedances.

    imaginary_residual: ndarray
        The residuals of the imaginary part of the complex impedances.

    mask: Dict[int, bool]
        The mask that was applied to the DataSet that the circuit was fitted to.

    chisqr: float
        The chi-squared value calculated for the result.

    red_chisqr: float
        The reduced chi-squared value calculated for the result.

    aic: float
        The calculated Akaike information criterion.

    bic: float
        The calculated Bayesian information criterion.

    ndata: int
        The number of data points.

    nfree: int
        The degrees of freedom.

    nfev: int
        The number of function evaluations.

    method: CNLSMethod
        The iterative method that produced the result.

    weight: Weight
        The weight function that produced the result.

    settings: FitSettings
        The settings that were used to perform the fit.
    """

    uuid: str
    timestamp: float
    circuit: Circuit
    parameters: Dict[str, Dict[str, FittedParameter]]
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray
    mask: Dict[int, bool]
    chisqr: float
    red_chisqr: float
    aic: float
    bic: float
    ndata: int
    nfree: int
    nfev: int
    method: CNLSMethod
    weight: Weight
    settings: FitSettings

    def __post_init__(self):
        self._cached_frequency: Dict[int, ndarray] = {}
        self._cached_impedance: Dict[int, ndarray] = {}

    def __repr__(self) -> str:
        return f"FitResult ({self.get_label()}, {hex(id(self))})"

    # TODO: Refactor
    @classmethod
    def from_dict(Class, dictionary: dict) -> "FitResult":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_result_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        mask: Dict[str, bool] = dictionary["mask"]
        if len(mask) < len(dictionary["frequency"]):
            i: int
            for i in range(0, len(dictionary["frequency"])):
                if mask.get(str(i)) is not True:
                    mask[str(i)] = False
        if (
            "real_impedance" not in dictionary
            or "imaginary_impedance" not in dictionary
        ):
            Z: ndarray = pyimpspec.parse_cdc(dictionary["circuit"]).impedances(
                dictionary["frequency"]
            )
            dictionary["real_impedance"] = list(Z.real)
            dictionary["imaginary_impedance"] = list(Z.imag)
        dictionary = parsers[version](dictionary)
        return Class(**dictionary)

    def to_dict(self, session: bool) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        assert type(session) is bool, session
        dictionary: dict = {
            "version": VERSION,
            "uuid": self.uuid,
            "timestamp": self.timestamp,
            "circuit": self.circuit.to_string(12),
            "parameters": {
                element_label: {
                    parameter_label: param.to_dict()
                    for parameter_label, param in parameters.items()
                }
                for element_label, parameters in self.parameters.items()
            },
            "frequency": list(self.frequency),
            "mask": {k: True for k, v in self.mask.items() if v is True},
            "real_residual": list(self.real_residual),
            "imaginary_residual": list(self.imaginary_residual),
            "chisqr": self.chisqr,
            "red_chisqr": self.red_chisqr,
            "aic": self.aic,
            "bic": self.bic,
            "ndata": self.ndata,
            "nfree": self.nfree,
            "nfev": self.nfev,
            "method": self.method,
            "weight": self.weight,
            "settings": self.settings.to_dict(),
        }
        if session:
            dictionary.update(
                {
                    "real_impedance": list(self.impedance.real),
                    "imaginary_impedance": list(self.impedance.imag),
                }
            )
        return dictionary

    def to_dataframe(self) -> DataFrame:
        """
        Get a `pandas.DataFrame` instance containing a table of fitted element parameters.
        """
        element_labels: List[str] = []
        parameter_labels: List[str] = []
        fitted_values: List[float] = []
        stderr_values: List[Union[float, str]] = []
        fixed: List[str] = []
        element_label: str
        parameters: Union[
            Dict[str, FittedParameter], Dict[int, OrderedDict[str, float]]
        ]
        element: Element
        for element in sorted(
            self.circuit.get_elements(flattened=True),
            key=lambda _: _.get_identifier(),
        ):
            element_label = element.get_label()
            parameters = self.parameters[element_label]
            parameter_label: str
            param: FittedParameter
            for parameter_label in sorted(parameters.keys()):
                param = parameters[parameter_label]
                element_labels.append(element_label)
                parameter_labels.append(parameter_label)
                fitted_values.append(param.value)
                stderr_values.append(
                    param.stderr / param.value * 100
                    if param.stderr is not None
                    else "-"
                )
                fixed.append("Yes" if param.fixed else "No")
        return DataFrame.from_dict(
            {
                "Element": element_labels,
                "Parameter": parameter_labels,
                "Value": fitted_values,
                "Std. err. (%)": stderr_values,
                "Fixed": fixed,
            }
        )

    def get_label(self) -> str:
        """
        Generate a label for the result.
        """
        cdc: str = self.settings.cdc
        while "{" in cdc:
            i: int = cdc.find("{")
            j: int = cdc.find("}")
            cdc = cdc.replace(cdc[i : j + 1], "")
        if cdc.startswith("[") and cdc.endswith("]"):
            cdc = cdc[1:-1]
        timestamp: str = format_timestamp(self.timestamp)
        return f"{cdc} ({timestamp})"

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        """
        Get an array of frequencies within the range of frequencies in the data set.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of fitted frequencies.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_frequency:
                self._cached_frequency.clear()
                self._cached_frequency[num_per_decade] = _interpolate(
                    self.frequency, num_per_decade
                )
            return self._cached_frequency[num_per_decade]
        return self.frequency

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        """
        Get the complex impedances produced by the fitted circuit within the range of frequencies in the data set.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of fitted frequencies and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_impedance:
                self._cached_impedance.clear()
                self._cached_impedance[num_per_decade] = self.circuit.impedances(
                    self.get_frequency(num_per_decade)
                )
            return self._cached_impedance[num_per_decade]
        return self.impedance

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        """
        Get the data required to plot the results as a Nyquist plot (-Z\" vs Z').

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of frequencies in the data set and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                Z.real,
                -Z.imag,
            )
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(
        self, num_per_decade: int = -1
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data required to plot the results as a Bode plot (|Z| and phi vs f).

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of frequencies in the data set and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            freq: ndarray = self.get_frequency(num_per_decade)
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                freq,
                abs(Z),
                -angle(Z, deg=True),
            )
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data required to plot the residuals (real and imaginary vs f).
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


general = SimpleNamespace(
    **{
        "apply_settings": """
Apply the settings that were used to generate the currently selected result.
    """.strip(),
        "apply_mask": """
Apply the mask that was used to generate the currently selected result.
    """.strip(),
        "copy_plot_data_as_csv": """
Copy the data in the plot to the clipboard as character-separated values.
    """.strip(),
        "copy_output": """
Copy the chosen output to the clipboard.
    """.strip(),
        "open_circuit_editor": """
Open the circuit editor.
    """.strip(),
        "adjust_nyquist_limits": """
Automatically adjust the limits of the Nyquist plot to fit the data.
    """.strip(),
        "adjust_bode_limits": """
Automatically adjust the limits of the Bode plot to fit the data.
    """.strip(),
        "adjust_residuals_limits": """
Automatically adjust the limits of the residuals plot to fit the data.
    """.strip(),
        "adjust_drt_limits": """
Automatically adjust the limits of the DRT plot to fit the data.
    """.strip(),
        "adjust_impedance_limits": """
Automatically adjust the limits of the impedance plot to fit the data.
    """.strip(),
        "adjust_limits": """
Automatically adjust the limits of the plots to fit the data.
    """.strip(),
        "command_palette": """
Search for an action to perform based on its description. Results can be navigated using the following keys:
- Home - first result
- Page up - five steps up
- Up arrow - one step up
- Down arrow - one step down
- Page down - five steps down
- End - last result

The highlighted result can be executed by pressing Enter. The window can be closed by pressing Esc.
    """.strip(),
        "auto_backup_interval": """
The number of actions between automatically saving a backup of the current state of a project in case of, e.g., crashes or power outages.

Setting this interval to zero disables automatic backups.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, List
from numpy import array, ndarray
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Nyquist
import deareis.themes as themes
from deareis.utility import calculate_window_position_dimensions
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import DataSet
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class CopyMask:
    def __init__(self, data: DataSet, data_sets: List[DataSet], callback: Callable):
        assert type(data) is DataSet, data
        assert type(data_sets) is list and all(
            map(lambda _: type(_) is DataSet, data_sets)
        ), data_sets
        self.data: DataSet = data
        self.preview_data: DataSet = DataSet.from_dict(data.to_dict())
        self.data_sets: List[DataSet] = [
            _ for _ in sorted(data_sets, key=lambda _: _.get_label()) if _ != data
        ]
        self.labels: List[str] = list(map(lambda _: _.get_label(), self.data_sets))
        self.callback: Callable = callback
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Copy mask",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            tag=self.window,
            on_close=self.close,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Source")
                attach_tooltip(tooltips.data_sets.copy_mask_source)
                self.combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    items=self.labels,
                    default_value=self.labels[0],
                    width=-1,
                    tag=self.combo,
                    callback=lambda s, a, u: self.select_source(a),
                )
            self.nyquist_plot: Nyquist = Nyquist(width=-1, height=-24)
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Excluded",
                theme=themes.nyquist.data,
            )
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Included",
                theme=themes.bode.phase_data,
                show_label=False,
            )
            self.nyquist_plot.plot(
                real=array([]),
                imaginary=array([]),
                label="Included",
                line=True,
                theme=themes.bode.phase_data,
            )
            dpg.add_button(label="Accept", callback=self.accept)
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.cycle_source(-1),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.cycle_source(1),
            )
        self.select_source(self.labels[0])
        self.nyquist_plot.queue_limits_adjustment()

    def cycle_source(self, step: int):
        assert type(step) is int, step
        index: int = self.labels.index(dpg.get_value(self.combo)) + step
        label: str = self.labels[index % len(self.labels)]
        dpg.set_value(self.combo, label)
        self.select_source(label)

    def select_source(self, label: str):
        assert type(label) is str, label
        self.preview_data.set_mask(self.data.get_mask())
        self.preview_data.set_mask(self.data_sets[self.labels.index(label)].get_mask())
        self.update_preview()

    def update_preview(self):
        real: ndarray
        imag: ndarray
        real, imag = self.preview_data.get_nyquist_data(masked=True)
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        real, imag = self.preview_data.get_nyquist_data(masked=False)
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )

    def close(self):
        dpg.hide_item(self.window)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def accept(self, keybinding: bool = False):
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        self.callback(
            self.data_sets[self.labels.index(dpg.get_value(self.combo))].get_mask(),
        )
        self.close()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Tuple,
)
from numpy import (
    angle,
    array,
    inf,
    integer,
    issubdtype,
    log10 as log,
    ndarray,
)
import pyimpspec
from pyimpspec import Circuit
from deareis.data.data_sets import DataSet
from pyimpspec.analysis.fitting import _interpolate
from deareis.enums import (
    CNLSMethod,
    TestMode,
    Test,
)
from deareis.utility import format_timestamp


VERSION: int = 1


def _parse_settings_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "test": Test(dictionary["test"]),
        "mode": TestMode(dictionary["mode"]),
        "num_RC": dictionary["num_RC"],
        "mu_criterion": dictionary["mu_criterion"],
        "add_capacitance": dictionary["add_capacitance"],
        "add_inductance": dictionary["add_inductance"],
        "method": CNLSMethod(dictionary["method"]),
        "max_nfev": dictionary["max_nfev"],
    }


@dataclass(frozen=True)
class TestSettings:
    """
    A class to store the settings used to perform a Kramers-Kronig test.

    Parameters
    ----------
    test: Test
        The type of test to perform: complex, real, imaginary, or CNLS.
        See pyimpspec and its documentation for details about the different types of tests.

    mode: TestMode
        How to perform the test: automatic, exploratory, or manual.
        The automatic mode uses the procedure described by Schönleber et al. (2014) to determine a suitable number of parallel RC circuits connected in series.
        The exploratory mode is similar to the automatic mode except the user is allowed to choose which of the results to accept and the initial suggestion has additional weighting applied to it in an effort to reduce false negatives that would lead to the conclusion that the data is invalid.
        The manual mode requires the user to pick the number of parallel RC circuits connected in series.

    num_RC: int
        The (maximum) number of parallel RC circuits connected in series.

    mu_criterion: float
        The threshold value used in the procedure described by Schönleber et al. (2014).

    add_capacitance: bool
        Add a capacitance in series to the Voigt circuit.

    add_inductance: bool
        Add an inductance in series to the Voigt circuit.

    method: CNLSMethod
        The iterative method to use if the CNLS test is chosen.

    max_nfev: int
        The maximum number of function evaluations to use if the CNLS test is chosen.
    """

    test: Test
    mode: TestMode
    num_RC: int
    mu_criterion: float
    add_capacitance: bool
    add_inductance: bool
    method: CNLSMethod
    max_nfev: int

    def __repr__(self) -> str:
        return f"TestSettings ({hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "TestSettings":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict, type(dictionary)
        assert "version" in dictionary
        version: int = dictionary["version"]
        del dictionary["version"]
        assert version >= 1 and version <= VERSION, (
            version,
            VERSION,
        )
        parsers: Dict[int, Callable] = {
            1: _parse_settings_v1,
        }
        assert version in parsers, (
            version,
            parsers,
        )
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        return Class(**parsers[version](dictionary))

    def to_dict(self) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        return {
            "version": VERSION,
            "test": self.test,
            "mode": self.mode,
            "num_RC": self.num_RC,
            "mu_criterion": self.mu_criterion,
            "add_capacitance": self.add_capacitance,
            "add_inductance": self.add_inductance,
            "method": self.method,
            "max_nfev": self.max_nfev,
        }


def _parse_result_v1(dictionary: dict) -> dict:
    assert type(dictionary) is dict
    return {
        "uuid": dictionary["uuid"],
        "timestamp": dictionary["timestamp"],
        "circuit": pyimpspec.parse_cdc(dictionary["circuit"]),
        "num_RC": dictionary["num_RC"],
        "mu": dictionary["mu"],
        "pseudo_chisqr": dictionary["pseudo_chisqr"],
        "frequency": array(dictionary["frequency"]),
        "real_residual": array(dictionary["real_residual"]),
        "imaginary_residual": array(dictionary["imaginary_residual"]),
        "mask": {int(k): v for k, v in dictionary.get("mask", {}).items()},
        "impedance": array(
            list(
                map(
                    lambda _: complex(*_),
                    zip(
                        dictionary["real_impedance"],
                        dictionary["imaginary_impedance"],
                    ),
                )
            )
        ),
        "settings": TestSettings.from_dict(dictionary["settings"]),
    }


@dataclass
class TestResult:
    """
    A class containing the result of a Kramers-Kronig test.

    Parameters
    ----------
    uuid: str
        The universally unique identifier assigned to this result.

    timestamp: float
        The Unix time (in seconds) for when the test was performed.

    circuit: Circuit
        The final, fitted circuit.

    num_RC: int
        The final number of parallel RC circuits connected in series.

    mu: float
        The mu-value that was calculated for the result.

    pseudo_chisqr: float
        The pseudo chi-squared value calculated according to eq. N in Boukamp (1995).

    frequency: ndarray
        The frequencies used to perform the test.

    impedance: ndarray
        The complex impedances of the fitted circuit at each of the frequencies.

    real_residual: ndarray
        The residuals of the real part of the complex impedances.

    imaginary_residual: ndarray
        The residuals of the imaginary part of the complex impedances.

    mask: Dict[int, bool]
        The mask that was applied to the DataSet that was tested.

    settings: TestSettings
        The settings that were used to perform the test.
    """

    uuid: str
    timestamp: float
    circuit: Circuit
    num_RC: int
    mu: float
    pseudo_chisqr: float
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray
    mask: Dict[int, bool]
    settings: TestSettings

    def __post_init__(self):
        self._cached_frequency: Dict[int, ndarray] = {}
        self._cached_impedance: Dict[int, ndarray] = {}

    def __repr__(self) -> str:
        return f"TestResult ({self.get_label()}, {hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "TestResult":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_result_v1,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        mask: Dict[str, bool] = dictionary["mask"]
        if len(mask) < len(dictionary["frequency"]):
            i: int
            for i in range(0, len(dictionary["frequency"])):
                if mask.get(str(i)) is not True:
                    mask[str(i)] = False
        if (
            "real_impedance" not in dictionary
            or "imaginary_impedance" not in dictionary
        ):
            Z: ndarray = pyimpspec.parse_cdc(dictionary["circuit"]).impedances(
                dictionary["frequency"]
            )
            dictionary["real_impedance"] = list(Z.real)
            dictionary["imaginary_impedance"] = list(Z.imag)
        return Class(**parsers[version](dictionary))

    def to_dict(self, session: bool) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        dictionary: dict = {
            "version": VERSION,
            "uuid": self.uuid,
            "timestamp": self.timestamp,
            "circuit": self.circuit.to_string(12),
            "num_RC": self.num_RC,
            "mu": self.mu,
            "pseudo_chisqr": self.pseudo_chisqr,
            "frequency": list(self.frequency),
            "real_residual": list(self.real_residual),
            "imaginary_residual": list(self.imaginary_residual),
            "mask": {k: True for k, v in self.mask.items() if v is True},
            "settings": self.settings.to_dict(),
        }
        if session:
            dictionary.update(
                {
                    "real_impedance": list(self.impedance.real),
                    "imaginary_impedance": list(self.impedance.imag),
                }
            )
        return dictionary

    def get_label(self) -> str:
        """
        Generate a label for the result.
        """
        circuit: str = f"R(RC){self.num_RC}"
        if self.settings.add_capacitance:
            circuit += "C"
        if self.settings.add_inductance:
            circuit += "L"
        timestamp: str = format_timestamp(self.timestamp)
        return f"{circuit} ({timestamp})"

    def get_frequency(self, num_per_decade: int = -1) -> ndarray:
        """
        Get an array of frequencies within the range of tested frequencies.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of tested frequencies.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_frequency:
                self._cached_frequency.clear()
                self._cached_frequency[num_per_decade] = _interpolate(
                    self.frequency, num_per_decade
                )
            return self._cached_frequency[num_per_decade]
        return self.frequency

    def get_impedance(self, num_per_decade: int = -1) -> ndarray:
        """
        Get the complex impedances produced by the fitted circuit within the range of tested frequencies.

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of tested frequencies and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            if num_per_decade not in self._cached_impedance:
                self._cached_impedance.clear()
                self._cached_impedance[num_per_decade] = self.circuit.impedances(
                    self.get_frequency(num_per_decade)
                )
            return self._cached_impedance[num_per_decade]
        return self.impedance

    def get_nyquist_data(self, num_per_decade: int = -1) -> Tuple[ndarray, ndarray]:
        """
        Get the data required to plot the results as a Nyquist plot (-Z\" vs Z').

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of tested frequencies and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                Z.real,
                -Z.imag,
            )
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(
        self, num_per_decade: int = -1
    ) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data required to plot the results as a Bode plot (|Z| and phi vs f).

        Parameters
        ----------
        num_per_decade: int = -1
            If the value is greater than zero, then logarithmically distributed frequencies will be generated within the range of tested frequencies and used to calculate the impedance produced by the fitted circuit.
        """
        assert issubdtype(type(num_per_decade), integer), num_per_decade
        if num_per_decade > 0:
            freq: ndarray = self.get_frequency(num_per_decade)
            Z: ndarray = self.get_impedance(num_per_decade)
            return (
                freq,
                abs(Z),
                -angle(Z, deg=True),
            )
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data required to plot the residuals (real and imaginary vs f).
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )

    def calculate_score(self, mu_criterion: float) -> float:
        """
        Calculate a score based on the provided mu-criterion and the statistics of the result.
        A result with a mu-value greater than or equal to the mu-criterion will get a score of -numpy.inf.

        Parameters
        ----------
        mu_criterion: float
            The mu-criterion to apply.
            See perform_test for details.

        Returns
        -------
        float
        """
        return (
            -inf
            if self.mu >= mu_criterion
            else -log(self.pseudo_chisqr) / (abs(mu_criterion - self.mu) ** 0.75)
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
import deareis.themes as themes


class OverviewTab:
    def __init__(self):
        self.tab: int = dpg.generate_uuid()
        self.label_input: int = dpg.generate_uuid()
        self.notes_input: int = dpg.generate_uuid()
        with dpg.tab(label="Overview", tag=self.tab):
            with dpg.group(horizontal=True):
                dpg.add_text("Label")
                dpg.add_input_text(
                    on_enter=True,
                    callback=lambda s, a, u: signals.emit(
                        Signal.RENAME_PROJECT, label=a
                    ),
                    width=-1,
                    tag=self.label_input,
                )
            dpg.add_text("Notes")
            dpg.add_input_text(
                multiline=True,
                tab_input=True,
                callback=lambda s, a, u: signals.emit(
                    Signal.MODIFY_PROJECT_NOTES, timers=u
                ),
                user_data=[],
                width=-1,
                height=-1,
                tag=self.notes_input,
            )

    def resize(self, width: int, height: int):
        pass

    def set_label(self, label: str):
        assert type(label) is str, label
        dpg.set_value(self.label_input, label)

    def set_notes(self, notes: str):
        assert type(notes) is str, notes
        dpg.set_value(self.notes_input, notes)

    def get_notes(self) -> str:
        return dpg.get_value(self.notes_input)

    def has_active_input(self) -> bool:
        return dpg.is_item_active(self.label_input) or dpg.is_item_active(
            self.notes_input
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace
from deareis.tooltips.kramers_kronig import boukamp1995


fitting = SimpleNamespace(
    **{
        "cdc": """
The equivalent circuit to fit to the data. A circuit description code (CDC) can be typed in to define the equivalent circuit. Alternatively, the graphical circuit editor can be used to construct the equivalent circuit.
    """.strip(),
        "method": """
See the documentation for the 'lmfit' Python package for information about the fitting methods.

The 'Auto' setting performs parallel fits with each of the fitting methods and returns the result with the smallest chi-squared value.
    """.strip(),
        "weight": f"""
The 'Auto' setting performs parallel fits with each of the weight functions listed above and returns the results with the smallest chi-squared value.

Reference: {boukamp1995}
    """.strip(),
        "nfev": """
The maximum number of function evaluations to perform. This setting can be used to limit the amount of time spent performing a fit. If this setting is set to zero, then no limit is applied.
    """.strip(),
        "perform": """
Sometimes a circuit, which ostensibly should be capable of providing a good fit, does not fit well. If the starting values of the parameters have been left at their default values, then rough adjustments may be all that is required. For example, if a circuit contains two parallel RC subcircuits in series (i.e. the CDC contains '(RC)(RC)'), then setting one of the capacitances to be one order of magnitude greater than the other may be sufficient.

Setting both the fitting method and weight function to 'Auto' is recommended as the starting point. Some of the various combinations of fitting methods and weight functions may be more forgiving than others regarding starting values for a particular circuit and thus require less work on the part of the user to achieve a good fit.

The fitting is performed using the Python package called lmfit. That package provides estimates for the error of the fitted parameters whenever possible. Circumstances when it is not possible to provide such estimates include e.g. when a parameter is near its lower/upper limit if such a limit has been defined. A parameter with a large error estimate is indicative of that parameter not really affecting the fit at all. For example, an unnecessary Warburg diffusion element may have its admittance parameter, Y, set to be a very large value and thus the element contributes very little to the overall impedance of the circuit within the frequency range of the experimental data.
    """.strip(),
        "chisqr": """
The chi-squared value of the fit.
    """.strip(),
        "red_chisqr": """
The reduced chi-squared value of the fit.
    """.strip(),
        "aic": """
The Akaike information criterion.
    """.strip(),
        "bic": """
The Bayesian information criterion.
    """.strip(),
        "nfree": """
The degrees of freedom in the fit.
    """.strip(),
        "ndata": """
The number of data points in the fit.
    """.strip(),
        "element": """
An element in the equivalent circuit.
    """.strip(),
        "parameter": """
A parameter in the element.
    """.strip(),
        "parameter_value": """
The fitted value of the parameter.
    """.strip(),
        "error": """
The estimated standard error of the fitted value.
    """.strip(),
        "delete": """
Delete the current fit result.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from math import floor
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
)
import dearpygui.dearpygui as dpg
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg
from numpy import asarray, float32
from deareis.data import (
    PlotSettings,
    Project,
)
from deareis.data.plotting import PlotExportSettings
from deareis.gui.plots import Image
from deareis.signals import Signal
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
from deareis.utility import format_number
from deareis.utility import calculate_window_position_dimensions
import deareis.api.plot.mpl as mpl
import deareis.signals as signals
import deareis.tooltips as tooltips
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)
from deareis.config import Config
from deareis.enums import (
    PLOT_EXTENSIONS,
    PlotPreviewLimit,
    plot_legend_location_to_label,
    label_to_plot_legend_location,
    plot_units_to_label,
    plot_units_per_inch,
    plot_preview_limit_to_label,
    label_to_plot_preview_limit,
    label_to_plot_units,
)


class SettingsMenu:
    def __init__(
        self,
        default_settings: PlotExportSettings,
        label_pad: int,
        refresh_callback: Optional[Callable] = None,
    ):
        callback_kwarg: dict = {}
        if refresh_callback is not None:
            callback_kwarg.update(
                {"callback": lambda s, a, u: refresh_callback(self.get_settings())}
            )
        with dpg.collapsing_header(label=" Dimensions", leaf=True):
            with dpg.group(horizontal=True):
                dpg.add_text("Units".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_units)
                self.unit_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    width=-1,
                    default_value=plot_units_to_label[default_settings.units],
                    items=list(label_to_plot_units.keys()),
                    **callback_kwarg,
                    tag=self.unit_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Width".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_width)
                self.width_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    default_value=default_settings.width,
                    min_value=0.01,
                    min_clamped=True,
                    step=0.0,
                    format="%.2f",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.width_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Height".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_height)
                self.height_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    default_value=default_settings.height,
                    min_value=0.01,
                    min_clamped=True,
                    step=0.0,
                    format="%.2f",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.height_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("DPI".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_dpi)
                self.dpi_input: int = dpg.generate_uuid()
                dpg.add_input_int(
                    width=-1,
                    default_value=default_settings.dpi,
                    min_value=1,
                    min_clamped=True,
                    step=0,
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.dpi_input,
                )
            if refresh_callback is not None:
                with dpg.group(horizontal=True):
                    dpg.add_text("Result".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_result)
                    tooltip_tag: int = dpg.generate_uuid()
                    self.dimensions_output: int = dpg.add_text(
                        "", user_data=tooltip_tag
                    )
                    attach_tooltip("", tag=tooltip_tag)
            with dpg.group(horizontal=True):
                dpg.add_text("Preview".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_preview)
                self.preview_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    width=-1,
                    default_value=plot_preview_limit_to_label[
                        default_settings.preview_limit
                    ],
                    items=list(label_to_plot_preview_limit.keys()),
                    **callback_kwarg,
                    tag=self.preview_combo,
                )
        with dpg.collapsing_header(
            label=" Limits",
            leaf=True,
            show=refresh_callback is not None,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Min. X".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_min_x)
                self.x_min_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=False,
                    **callback_kwarg,
                    tag=self.x_min_checkbox,
                )
                self.x_min_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    step=0.0,
                    format="%.2e",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.x_min_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Max. X".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_max_x)
                self.x_max_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=False,
                    **callback_kwarg,
                    tag=self.x_max_checkbox,
                )
                self.x_max_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    step=0.0,
                    format="%.2e",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.x_max_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Min. Y".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_min_y)
                self.y_min_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=False,
                    **callback_kwarg,
                    tag=self.y_min_checkbox,
                )
                self.y_min_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    step=0.0,
                    format="%.2e",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.y_min_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Max. Y".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_max_y)
                self.y_max_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=False,
                    **callback_kwarg,
                    tag=self.y_max_checkbox,
                )
                self.y_max_input: int = dpg.generate_uuid()
                dpg.add_input_float(
                    width=-1,
                    step=0.0,
                    format="%.2e",
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.y_max_input,
                )
        with dpg.collapsing_header(label=" Layout", leaf=True):
            with dpg.group(horizontal=True):
                dpg.add_text("Title".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_title)
                self.show_title_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.show_title,
                    **callback_kwarg,
                    tag=self.show_title_checkbox,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Legend".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_legend)
                self.show_legend_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.show_legend,
                    **callback_kwarg,
                    tag=self.show_legend_checkbox,
                )
                self.legend_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=plot_legend_location_to_label[
                        default_settings.legend_location
                    ],
                    items=list(label_to_plot_legend_location.keys()),
                    **callback_kwarg,
                    width=-1,
                    tag=self.legend_combo,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Grid".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_grid)
                self.show_grid_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.show_grid,
                    **callback_kwarg,
                    tag=self.show_grid_checkbox,
                )
            with dpg.group(horizontal=True):
                dpg.add_text("Tight".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_tight)
                self.has_tight_layout_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.has_tight_layout,
                    **callback_kwarg,
                    tag=self.has_tight_layout_checkbox,
                )
        with dpg.collapsing_header(label=" Miscellaneous", leaf=True):
            with dpg.group(horizontal=True):
                dpg.add_text("Points per decade".rjust(label_pad))
                attach_tooltip(tooltips.plotting.export_num_per_decade)
                self.num_per_decade_input: int = dpg.generate_uuid()
                dpg.add_input_int(
                    width=-1,
                    default_value=default_settings.num_per_decade,
                    min_value=1,
                    min_clamped=True,
                    step=0,
                    on_enter=True,
                    **callback_kwarg,
                    tag=self.num_per_decade_input,
                )
            with dpg.group(horizontal=True, show=refresh_callback is None):
                dpg.add_text("Default extension".rjust(label_pad))
                self.extension_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    width=-1,
                    default_value=default_settings.extension
                    if default_settings.extension in PLOT_EXTENSIONS
                    else ".png",
                    items=PLOT_EXTENSIONS,
                    tag=self.extension_combo,
                )
            with dpg.group(horizontal=True, show=refresh_callback is None):
                dpg.add_text("Clear texture registry".rjust(label_pad))
                attach_tooltip(tooltips.plotting.clear_registry)
                self.clear_registry_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.clear_registry,
                    tag=self.clear_registry_checkbox,
                )
            with dpg.group(horizontal=True, show=refresh_callback is None):
                dpg.add_text("Disable plot preview".rjust(label_pad))
                attach_tooltip(tooltips.plotting.disable_preview)
                self.disable_preview_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    default_value=default_settings.disable_preview,
                    tag=self.disable_preview_checkbox,
                )

    def get_settings(self) -> PlotExportSettings:
        return PlotExportSettings(
            units=label_to_plot_units[dpg.get_value(self.unit_combo)],
            width=dpg.get_value(self.width_input),
            height=dpg.get_value(self.height_input),
            dpi=dpg.get_value(self.dpi_input),
            preview_limit=label_to_plot_preview_limit[
                dpg.get_value(self.preview_combo)
            ],
            show_title=dpg.get_value(self.show_title_checkbox),
            show_legend=dpg.get_value(self.show_legend_checkbox),
            legend_location=label_to_plot_legend_location[
                dpg.get_value(self.legend_combo)
            ],
            show_grid=dpg.get_value(self.show_grid_checkbox),
            has_tight_layout=dpg.get_value(self.has_tight_layout_checkbox),
            num_per_decade=dpg.get_value(self.num_per_decade_input),
            extension=dpg.get_value(self.extension_combo),
            clear_registry=dpg.get_value(self.clear_registry_checkbox),
            disable_preview=dpg.get_value(self.disable_preview_checkbox),
        )

    def set_settings(self, settings: PlotExportSettings):
        dpg.set_value(self.unit_combo, plot_units_to_label[settings.units])
        dpg.set_value(self.width_input, settings.width)
        dpg.set_value(self.height_input, settings.height)
        dpg.set_value(self.dpi_input, settings.dpi)
        dpg.set_value(
            self.preview_combo,
            plot_preview_limit_to_label[settings.preview_limit],
        )
        dpg.set_value(self.show_title_checkbox, settings.show_title)
        dpg.set_value(self.show_legend_checkbox, settings.show_legend)
        if dpg.get_value(self.show_legend_checkbox):
            dpg.enable_item(self.legend_combo)
        else:
            dpg.disable_item(self.legend_combo)
        dpg.set_value(
            self.legend_combo,
            plot_legend_location_to_label[settings.legend_location],
        )
        dpg.set_value(self.show_grid_checkbox, settings.show_grid)
        dpg.set_value(self.has_tight_layout_checkbox, settings.has_tight_layout)
        dpg.set_value(self.num_per_decade_input, settings.num_per_decade)
        dpg.set_value(self.extension_combo, settings.extension)
        dpg.set_value(self.clear_registry_checkbox, settings.clear_registry)
        dpg.set_value(self.disable_preview_checkbox, settings.disable_preview)
        if dpg.get_value(self.x_min_checkbox):
            dpg.enable_item(self.x_min_input)
        else:
            dpg.disable_item(self.x_min_input)
        if dpg.get_value(self.x_max_checkbox):
            dpg.enable_item(self.x_max_input)
        else:
            dpg.disable_item(self.x_max_input)
        if dpg.get_value(self.y_min_checkbox):
            dpg.enable_item(self.y_min_input)
        else:
            dpg.disable_item(self.y_min_input)
        if dpg.get_value(self.y_max_checkbox):
            dpg.enable_item(self.y_max_input)
        else:
            dpg.disable_item(self.y_max_input)

    def update_plot_limits(
        self, x_min: float, x_max: float, y_min: float, y_max: float
    ):
        if not dpg.get_value(self.x_min_checkbox):
            dpg.set_value(self.x_min_input, x_min)
        if not dpg.get_value(self.x_max_checkbox):
            dpg.set_value(self.x_max_input, x_max)
        if not dpg.get_value(self.y_min_checkbox):
            dpg.set_value(self.y_min_input, y_min)
        if not dpg.get_value(self.y_max_checkbox):
            dpg.set_value(self.y_max_input, y_max)

    def get_x_limits(self) -> Optional[Tuple[Optional[float], Optional[float]]]:
        if dpg.get_value(self.x_min_checkbox) or dpg.get_value(self.x_max_checkbox):
            return (
                dpg.get_value(self.x_min_input)
                if dpg.get_value(self.x_min_checkbox)
                else None,
                dpg.get_value(self.x_max_input)
                if dpg.get_value(self.x_max_checkbox)
                else None,
            )
        return None

    def get_y_limits(self) -> Optional[Tuple[Optional[float], Optional[float]]]:
        if dpg.get_value(self.y_min_checkbox) or dpg.get_value(self.y_max_checkbox):
            return (
                dpg.get_value(self.y_min_input)
                if dpg.get_value(self.y_min_checkbox)
                else None,
                dpg.get_value(self.y_max_input)
                if dpg.get_value(self.y_max_checkbox)
                else None,
            )
        return None

    def update_dimensions_output(self, pixel_dimensions: List[str]):
        dpg.set_value(
            self.dimensions_output,
            f"{pixel_dimensions[0]} px * {pixel_dimensions[1]} px",
        )
        update_tooltip(
            dpg.get_item_user_data(self.dimensions_output),
            f"{pixel_dimensions[0]} px * {pixel_dimensions[1]} px",
        )


class PlotExporter:
    def __init__(self, config: Config):
        self.key_handler: int = -1
        self.settings: Optional[PlotSettings] = None
        self.project: Optional[Project] = None
        self.texture_registry: int = dpg.generate_uuid()
        dpg.add_texture_registry(tag=self.texture_registry)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Export plot",
            show=False,
            modal=True,
            on_close=self.close,
            tag=self.window,
        ):
            with dpg.group(horizontal=True):
                dpg.add_spacer(width=8)
                self.sidebar: int = dpg.generate_uuid()
                with dpg.child_window(
                    width=250,
                    height=-1,
                    border=False,
                    tag=self.sidebar,
                ):
                    dpg.add_spacer(height=4)
                    with dpg.child_window(
                        border=False,
                        height=-32,
                    ):
                        self.settings_menu: SettingsMenu = SettingsMenu(
                            config.default_plot_export_settings,
                            label_pad=10,
                            refresh_callback=self.refresh,
                        )
                    with dpg.child_window(border=False):
                        self.save_button: int = dpg.add_button(
                            label="Save as",
                            callback=lambda s, a, u: self.save(fig=u),
                            user_data=None,
                            width=-1,
                        )
                        attach_tooltip(tooltips.plotting.export_file)
                self.image_plot: Image = Image()

    def clear(self):
        settings: PlotExportSettings = self.settings_menu.get_settings()
        fig: Optional[Figure] = dpg.get_item_user_data(self.save_button)
        if fig is not None:
            plt.close(fig)
            dpg.set_item_user_data(self.save_button, None)
        if not settings.disable_preview:
            self.image_plot.clear()
            if settings.clear_registry:
                dpg.delete_item(self.texture_registry, children_only=True)

    def set_settings(self, settings: PlotExportSettings):
        self.settings_menu.set_settings(settings)

    def close(self):
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        dpg.hide_item(self.window)
        self.clear()
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def initialize_keybindings(self):
        self.key_handler = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.save(
                    dpg.get_item_user_data(self.save_button), keybinding=True
                ),
            )

    def show(self, settings: PlotSettings, project: Project):
        self.initialize_keybindings()
        self.settings = settings
        self.project = project
        dpg.show_item(self.window)
        dpg.split_frame()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window, window_object=self)
        self.refresh(self.settings_menu.get_settings())

    def create_preview_figure(
        self,
        width: float,
        height: float,
        dpi: int,
        preview_limit: int,
    ) -> Tuple[Figure, int, int]:
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        pixel_width: int = floor(width * dpi)
        pixel_height: int = floor(height * dpi)
        if preview_limit > 0 and (
            pixel_width > preview_limit or pixel_height > preview_limit
        ):
            if pixel_width >= pixel_height:
                pixel_height = floor(pixel_height / pixel_width * preview_limit)
                pixel_width = preview_limit
            else:
                pixel_width = floor(pixel_width / pixel_height * preview_limit)
                pixel_height = preview_limit
        width = pixel_width / dpi
        height = pixel_height / dpi
        fig: Figure = plt.figure(
            figsize=(
                width,
                height,
            ),
            dpi=dpi,
        )
        return (
            fig,
            pixel_width,
            pixel_height,
        )

    def create_final_figure(self, width: float, height: float, dpi: int) -> Figure:
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        pixel_width: int = floor(width * dpi)
        pixel_height: int = floor(height * dpi)
        pixel_dimensions: List[str] = [
            format_number(float(pixel_width), decimals=0, exponent=False),
            format_number(float(pixel_height), decimals=0, exponent=False),
        ]
        self.settings_menu.update_dimensions_output(pixel_dimensions)
        fig: Figure = plt.figure(
            figsize=(
                width,
                height,
            ),
            dpi=dpi,
        )
        return fig

    def plot(self, fig: Figure, export_settings: PlotExportSettings):
        assert type(fig) is Figure, type(fig)
        assert type(export_settings) is PlotExportSettings, type(export_settings)
        assert self.settings is not None
        assert self.project is not None
        x_limits: Optional[
            Tuple[Optional[float], Optional[float]]
        ] = self.settings_menu.get_x_limits()
        y_limits: Optional[
            Tuple[Optional[float], Optional[float]]
        ] = self.settings_menu.get_y_limits()
        mpl.plot(
            self.settings,
            self.project,
            x_limits=x_limits,
            y_limits=y_limits,
            show_title=export_settings.show_title,
            show_legend=export_settings.show_legend,
            legend_loc=int(export_settings.legend_location),
            tight_layout=export_settings.has_tight_layout,
            show_grid=export_settings.show_grid,
            fig=fig,
            axis=fig.gca(),
            num_per_decade=export_settings.num_per_decade,
        )

    def refresh(self, settings: PlotExportSettings):
        self.settings_menu.set_settings(settings)
        width: float = settings.width
        height: float = settings.height
        dpi: int = settings.dpi
        upi: float = plot_units_per_inch[settings.units]
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        assert type(upi) is float and upi > 0.0, upi
        self.clear()
        width = width / upi
        height = height / upi
        final_fig: Figure = self.create_final_figure(width, height, dpi)
        dpg.set_item_user_data(self.save_button, final_fig)
        self.plot(final_fig, settings)
        # Preview
        if not settings.disable_preview:
            preview_fig: Figure
            pixel_width: int
            pixel_height: int
            preview_fig, pixel_width, pixel_height = self.create_preview_figure(
                width,
                height,
                dpi,
                2 ** int(settings.preview_limit)
                if settings.preview_limit != PlotPreviewLimit.NONE
                else 0,
            )
            canvas: FigureCanvasAgg = FigureCanvasAgg(preview_fig)
            self.plot(preview_fig, settings)
            canvas.draw()
            tag: int = dpg.add_raw_texture(
                pixel_width,
                pixel_height,
                asarray(canvas.buffer_rgba()).astype(float32) / 255,
                format=dpg.mvFormat_Float_rgba,
                parent=self.texture_registry,
            )
            self.image_plot.plot(
                texture=tag,
                bounds_min=(
                    0,
                    0,
                ),
                bounds_max=(
                    pixel_width,
                    pixel_height,
                ),
            )
            plt.close(preview_fig)
            self.image_plot.queue_limits_adjustment()
        x_min: float
        x_max: float
        y_min: float
        y_max: float
        x_min, x_max = final_fig.gca().get_xlim()
        y_min, y_max = final_fig.gca().get_ylim()
        self.settings_menu.update_plot_limits(x_min, x_max, y_min, y_max)

    def save(self, fig: Optional[Figure], keybinding: bool = False):
        if fig is None:
            return
        elif keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        extension: str = self.settings_menu.get_settings().extension
        self.close()
        dpg.split_frame()
        signals.emit(
            Signal.SAVE_PLOT,
            figure=fig,
            default_extension=extension if extension in PLOT_EXTENSIONS else ".png",
            extensions=PLOT_EXTENSIONS,
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    List,
    Tuple,
)
import dearpygui.dearpygui as dpg
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class Image(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            equal_aspects=True,
            no_mouse_pos=True,
            no_menus=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="",
                no_tick_labels=True,
                no_tick_marks=True,
                no_gridlines=True,
            )
            self._y_axis: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="",
                no_tick_labels=True,
                no_tick_marks=True,
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    def is_blank(self) -> bool:
        return len(dpg.get_item_children(self._y_axis, slot=1)) == 0

    def clear(self, *args, **kwargs):
        dpg.delete_item(self._y_axis, children_only=True)
        self._series.clear()

    def plot(self, *args, **kwargs) -> int:
        assert len(args) == 0, args
        texture: int = kwargs["texture"]
        bounds_min: Tuple[int, int] = kwargs["bounds_min"]
        bounds_max: Tuple[int, int] = kwargs["bounds_max"]
        assert type(texture) is int and texture > 0, texture
        assert type(bounds_min) is tuple and all(
            map(lambda _: type(_) is int, bounds_min)
        ), bounds_min
        assert type(bounds_max) is tuple and all(
            map(lambda _: type(_) is int, bounds_max)
        ), bounds_max
        self._series.append(kwargs)
        tag: int = dpg.add_image_series(
            texture,
            bounds_min,
            bounds_max,
            parent=self._y_axis,
        )
        return tag

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        dpg.split_frame()
        dpg.fit_axis_data(self._x_axis)
        dpg.fit_axis_data(self._y_axis)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis,
            ],
            [
                self._x_axis,
                self._y_axis,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis)
